# Copyright (C) 2020 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; If not, see <http://www.gnu.org/licenses/>.

from nbxmpp.protocol import Iq
from nbxmpp.protocol import NodeProcessed
from nbxmpp.namespaces import Namespace
from nbxmpp.structs import StanzaHandler
from nbxmpp.task import iq_request_task
from nbxmpp.modules.base import BaseModule
from nbxmpp.modules.util import process_response


class Ping(BaseModule):
    def __init__(self, client):
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = [
            StanzaHandler(name='iq',
                          callback=self._process_ping,
                          typ='get',
                          ns=Namespace.PING,
                          priority=15),
        ]

    def _process_ping(self, _client, stanza, _properties):
        self._log.info('Send pong to %s', stanza.getFrom())
        iq = stanza.buildSimpleReply('result')
        self._client.send_stanza(iq)
        raise NodeProcessed

    @iq_request_task
    def ping(self, jid):
        _task = yield

        response = yield _make_ping_request(jid)
        yield process_response(response)


def _make_ping_request(jid):
    iq = Iq('get', to=jid)
    iq.addChild(name='ping', namespace=Namespace.PING)
    return iq



# Copyright (C) 2021 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; If not, see <http://www.gnu.org/licenses/>.

from nbxmpp.protocol import Iq
from nbxmpp.protocol import NodeProcessed
from nbxmpp.protocol import Error
from nbxmpp.protocol import ERR_FORBIDDEN
from nbxmpp.protocol import ERR_SERVICE_UNAVAILABLE
from nbxmpp.namespaces import Namespace
from nbxmpp.task import iq_request_task
from nbxmpp.structs import StanzaHandler
from nbxmpp.errors import MalformedStanzaError
from nbxmpp.errors import StanzaError
from nbxmpp.modules.base import BaseModule

from nbxmpp.modules.date_and_time import parse_datetime
from nbxmpp.modules.date_and_time import create_tzinfo
from nbxmpp.modules.date_and_time import get_local_time


class EntityTime(BaseModule):
    def __init__(self, client):
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = [
            StanzaHandler(name='iq',
                          callback=self._answer_request,
                          priority=60,
                          typ='get',
                          ns=Namespace.TIME),
        ]

        self._enabled = False
        self._allow_reply_func = None

    def disable(self):
        self._enabled = False

    def enable(self):
        self._enabled = True

    def set_allow_reply_func(self, func):
        self._allow_reply_func = func

    @iq_request_task
    def request_entity_time(self, jid):
        _task = yield

        response = yield _make_request(jid)
        if response.isError():
            raise StanzaError(response)

        yield _parse_response(response)

    def _answer_request(self, _con, stanza, _properties):
        self._log.info('Request received from %s', stanza.getFrom())
        if not self._enabled:
            self._client.send_stanza(Error(stanza, ERR_SERVICE_UNAVAILABLE))
            raise NodeProcessed

        if self._allow_reply_func is not None:
            if not self._allow_reply_func(stanza.getFrom()):
                self._client.send_stanza(Error(stanza, ERR_FORBIDDEN))
                raise NodeProcessed

        time, tzo = get_local_time()
        iq = stanza.buildSimpleReply('result')
        time_node = iq.addChild('time', namespace=Namespace.TIME)
        time_node.setTagData('utc', time)
        time_node.setTagData('tzo', tzo)
        self._log.info('Send time: %s %s', time, tzo)
        self._client.send_stanza(iq)
        raise NodeProcessed


def _make_request(jid):
    iq = Iq('get', to=jid)
    iq.addChild('time', namespace=Namespace.TIME)
    return iq


def _parse_response(response):
    time_ = response.getTag('time')
    if not time_:
        raise MalformedStanzaError('time node missing', response)

    tzo = time_.getTagData('tzo')
    if not tzo:
        raise MalformedStanzaError('tzo node or data missing', response)

    remote_tz = create_tzinfo(tz_string=tzo)
    if remote_tz is None:
        raise MalformedStanzaError('invalid tzo data', response)

    utc_time = time_.getTagData('utc')
    if not utc_time:
        raise MalformedStanzaError('utc node or data missing', response)

    date_time = parse_datetime(utc_time, check_utc=True)
    if date_time is None:
        raise MalformedStanzaError('invalid timezone definition', response)

    date_time = date_time.astimezone(remote_tz)
    return date_time.strftime('%c %Z')



# Copyright (C) 2018 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; If not, see <http://www.gnu.org/licenses/>.

from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Node
from nbxmpp.protocol import NodeProcessed
from nbxmpp.structs import StanzaHandler
from nbxmpp.structs import MoodData
from nbxmpp.const import MOODS
from nbxmpp.modules.base import BaseModule
from nbxmpp.modules.util import finalize
from nbxmpp.task import iq_request_task


class Mood(BaseModule):

    _depends = {
        'publish': 'PubSub'
    }

    def __init__(self, client):
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = [
            StanzaHandler(name='message',
                          callback=self._process_pubsub_mood,
                          ns=Namespace.PUBSUB_EVENT,
                          priority=16),
        ]

    def _process_pubsub_mood(self, _client, stanza, properties):
        if not properties.is_pubsub_event:
            return

        if properties.pubsub_event.node != Namespace.MOOD:
            return

        item = properties.pubsub_event.item
        if item is None:
            # Retract, Deleted or Purged
            return

        mood_node = item.getTag('mood', namespace=Namespace.MOOD)
        if not mood_node.getChildren():
            self._log.info('Received mood: %s - removed mood', properties.jid)
            return

        mood, text = None, None
        for child in mood_node.getChildren():
            name = child.getName().strip()
            if name == 'text':
                text = child.getData()
            elif name in MOODS:
                mood = name

        if mood is None and mood_node.getChildren():
            self._log.warning('No valid mood value found')
            self._log.warning(stanza)
            raise NodeProcessed

        data = MoodData(mood, text)
        pubsub_event = properties.pubsub_event._replace(data=data)
        self._log.info('Received mood: %s - %s', properties.jid, data)

        properties.pubsub_event = pubsub_event

    @iq_request_task
    def set_mood(self, data):
        task = yield

        item = Node('mood', {'xmlns': Namespace.MOOD})
        if data is not None and data.mood:
            item.addChild(data.mood)

            if data.text:
                item.addChild('text', payload=data.text)

        result = yield self.publish(Namespace.MOOD, item, id_='current')

        yield finalize(task, result)



# Copyright (C) 2022 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; If not, see <http://www.gnu.org/licenses/>.
import typing

from nbxmpp.namespaces import Namespace
from nbxmpp.structs import StanzaHandler, MessageProperties
from nbxmpp.structs import Reactions as ReactionStruct
from nbxmpp.modules.base import BaseModule

if typing.TYPE_CHECKING:
    from nbxmpp.client import Client
    from nbxmpp.protocol import Message


class Reactions(BaseModule):
    def __init__(self, client: 'Client'):
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = [
            StanzaHandler(
                name='message',
                callback=self._process_message_reaction,
                ns=Namespace.REACTIONS,
                priority=15,
            ),
        ]

    def _process_message_reaction(
        self,
        _client: 'Client',
        stanza: 'Message',
        properties: MessageProperties
    ) -> None:
        reactions = stanza.getTag('reactions', namespace=Namespace.REACTIONS)

        if reactions is None:
            return

        id_ = reactions.getAttr('id')
        if not id_:
            self._log.warning('Reactions without ID')
            return

        emojis = set()
        for reaction in reactions.getTags('reaction'):
            # we strip for clients that might add white spaces and/or
            # new lines in the reaction content.
            emoji = reaction.getData().strip()
            if emoji:
                emojis.add(emoji)
            else:
                self._log.warning('Empty reaction')

        properties.reactions = ReactionStruct(id_, emojis)



# Copyright (C) 2020 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; If not, see <http://www.gnu.org/licenses/>.

from nbxmpp.namespaces import Namespace
from nbxmpp.structs import RSMData


def parse_rsm(stanza):
    stanza = stanza.getTag('set', namespace=Namespace.RSM)
    if stanza is None:
        return None

    after = stanza.getTagData('after') or None
    before = stanza.getTagData('before') or None
    last = stanza.getTagData('last') or None

    first_index = None
    first = stanza.getTagData('first') or None
    if first is not None:
        try:
            first_index = int(first.getAttr('index'))
        except Exception:
            pass

    try:
        count = int(stanza.getTagData('count'))
    except Exception:
        count = None

    try:
        max_ = int(stanza.getTagData('max'))
    except Exception:
        max_ = None

    try:
        index = int(stanza.getTagData('index'))
    except Exception:
        index = None

    return RSMData(after=after,
                   before=before,
                   last=last,
                   first=first,
                   first_index=first_index,
                   count=count,
                   max=max_,
                   index=index)



# Copyright (C) 2020 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; If not, see <http://www.gnu.org/licenses/>.

import logging
from collections import deque

from gi.repository import GLib
from gi.repository import Gio
from gi.repository import GObject

from nbxmpp.const import TCPState
from nbxmpp.const import ConnectionType
from nbxmpp.util import utf8_decode
from nbxmpp.util import convert_tls_error_flags
from nbxmpp.connection import Connection

log = logging.getLogger('nbxmpp.tcp')

READ_BUFFER_SIZE = 8192


class TCPConnection(Connection):
    def __init__(self, *args, **kwargs):
        Connection.__init__(self, *args, **kwargs)

        self._client = Gio.SocketClient.new()
        self._client.set_timeout(7)

        if self._address.proxy is not None:
            self._proxy_resolver = self._address.proxy.get_resolver()
            self._client.set_proxy_resolver(self._proxy_resolver)

        GObject.Object.connect(self._client, 'event', self._on_event)

        self._con = None

        self._read_buffer = b''

        self._write_queue = deque([])
        self._write_stanza_buffer = None

        self._connect_cancellable = Gio.Cancellable()
        self._read_cancellable = Gio.Cancellable()

        self._tls_handshake_in_progress = False
        self._input_closed = False
        self._output_closed = False

        self._keepalive_id = None

    def connect(self):
        self.state = TCPState.CONNECTING

        if self._address.is_service:
            self._client.connect_to_service_async(self._address.domain,
                                                  self._address.service,
                                                  self._connect_cancellable,
                                                  self._on_connect_finished,
                                                  None)
        elif self._address.is_host:
            self._client.connect_to_host_async(self._address.host,
                                               0,
                                               self._connect_cancellable,
                                               self._on_connect_finished,
                                               None)

        else:
            raise ValueError('Invalid Address')

    def _on_event(self, _socket_client, event, _connectable, connection):
        if event == Gio.SocketClientEvent.CONNECTING:
            self._remote_address = connection.get_remote_address().to_string()
            use_proxy = self._address.proxy is not None
            target = 'proxy' if use_proxy else self._address.domain
            self._log.info('Connecting to %s (%s)',
                           target,
                           self._remote_address)

    def _check_certificate(self, _connection, certificate, errors):
        self._peer_certificate = certificate
        self._peer_certificate_errors = convert_tls_error_flags(errors)

        if self._accept_certificate():
            return True

        self.notify('bad-certificate')
        return False

    def _on_certificate_set(self, connection, _param):
        if self._peer_certificate is None:
            # If the cert has errors _check_certificate() will set the cert and
            # _accept_certificate() will modify the error set. If this is the
            # case _accept_certificate() modifies the errors.
            self._peer_certificate = connection.props.peer_certificate
            self._peer_certificate_errors = convert_tls_error_flags(
                connection.props.peer_certificate_errors)

        self._tls_handshake_in_progress = False

    def _on_connect_finished(self, client, result, _user_data):
        try:
            if self._address.proxy is not None:
                self._con = client.connect_to_host_finish(result)
            elif self._address.is_service:
                self._con = client.connect_to_service_finish(result)
            elif self._address.is_host:
                self._con = client.connect_to_host_finish(result)
            else:
                raise ValueError('Address must be a service or host')
        except GLib.Error as error:
            self._log.info('Connect Error: %s', error)
            self._finalize('connection-failed')
            return

        # We use the timeout only for connecting
        self._con.get_socket().set_timeout(0)
        self._con.set_graceful_disconnect(True)
        self._con.get_socket().set_keepalive(True)

        self._local_address = self._con.get_local_address()

        self.state = TCPState.CONNECTED

        use_proxy = self._address.proxy is not None
        target = 'proxy' if use_proxy else self._address.domain
        self._log.info('Connected to %s (%s)',
                       target,
                       self._con.get_remote_address().to_string())

        self._on_connected()

    def _on_connected(self):
        self.notify('connected')
        self._read_async()

    def _remove_keepalive_timer(self):
        if self._keepalive_id is not None:
            self._log.info('Remove keepalive timer')
            GLib.source_remove(self._keepalive_id)
            self._keepalive_id = None

    def _renew_keepalive_timer(self):
        if self._con is None:
            return
        self._remove_keepalive_timer()
        self._log.info('Add keepalive timer')
        self._keepalive_id = GLib.timeout_add_seconds(5, self._send_keepalive)

    def _send_keepalive(self):
        self._log.info('Send keepalive')
        self._keepalive_id = None
        if not self._con.get_output_stream().has_pending():
            self._write_all_async(' '.encode())

    def start_tls_negotiation(self):
        self._log.info('Start TLS negotiation')
        self._tls_handshake_in_progress = True
        remote_address = self._con.get_remote_address()
        identity = Gio.NetworkAddress.new(self._address.domain,
                                          remote_address.props.port)

        self._tls_con = Gio.TlsClientConnection.new(self._con, identity)

        if self._address.type == ConnectionType.DIRECT_TLS:
            self._tls_con.set_advertised_protocols(['xmpp-client'])
        self._tls_con.connect('accept-certificate', self._check_certificate)
        self._tls_con.connect('notify::peer-certificate', self._on_certificate_set)

        # This Wraps the Gio.TlsClientConnection and the Gio.Socket together
        # so we get back a Gio.SocketConnection
        self._con = Gio.TcpWrapperConnection.new(self._tls_con,
                                                 self._con.get_socket())

    def _read_async(self):
        if self._input_closed:
            return

        self._con.get_input_stream().read_bytes_async(
            READ_BUFFER_SIZE,
            GLib.PRIORITY_LOW,
            self._read_cancellable,
            self._on_read_async_finish,
            None)

    def _on_read_async_finish(self, stream, result, _user_data):
        try:
            data = stream.read_bytes_finish(result)
        except GLib.Error as error:
            quark = GLib.quark_try_string('g-io-error-quark')
            if error.matches(quark, Gio.IOErrorEnum.CANCELLED):
                if self._input_closed:
                    return

            quark = GLib.quark_try_string('g-tls-error-quark')
            if error.matches(quark, Gio.TlsError.MISC):
                if self._tls_handshake_in_progress:
                    self._log.error('Handshake failed: %s', error)
                    self._finalize('connection-failed')
                    return

            if error.matches(quark, Gio.TlsError.EOF):
                self._log.info('Incoming stream closed: TLS EOF')
                self._finalize('disconnected')
                return

            if error.matches(quark, Gio.TlsError.BAD_CERTIFICATE):
                self._log.info('Certificate Error: %s', error)
                self._finalize('disconnected')
                return

            self._log.error('Read Error: %s', error)

            if self._state not in (TCPState.DISCONNECTING,
                                   TCPState.DISCONNECTED):
                self._finalize('disconnected')
            return

        except RuntimeError as error:
            # PyGObject raises a RuntimeError when it fails to convert the
            # GError. Why it failed is printed by PyGObject
            self._log.error(error)
            return

        data = data.get_data()
        if not data:
            self._log.info('Reveived zero data on _read_async()')
            self._finalize('disconnected')
            return

        self._renew_keepalive_timer()

        self._read_buffer += data

        try:
            data, self._read_buffer = utf8_decode(self._read_buffer)
        except UnicodeDecodeError as error:
            self._log.warning(error)
            self._log.warning('read buffer: "%s"', self._read_buffer)
            self._log.warning('data: "%s"', data)
            self._finalize('disconnected')
            return

        self._log_stanza(data, received=True)

        try:
            self.notify('data-received', data)
        except Exception:
            self._log.exception('Error while executing data-received:')

        # Call next async read only after the received data is processed
        # otherwise this can lead to problems if we call
        # start_tls_negotiation() while we have a pending read which is
        # the case for START TLS because its triggered by <proceed>
        self._read_async()

    def _write_stanzas(self):
        self._write_stanza_buffer = self._write_queue
        self._write_queue = deque([])
        data = ''.join(map(str, self._write_stanza_buffer)).encode()
        self._write_all_async(data)

    def _write_all_async(self, data):
        # We have to pass data to the callback, because GLib takes no
        # reference on the passed data and python would gc collect it
        # bevor GLib has a chance to write it to the stream
        self._con.get_output_stream().write_all_async(
            data,
            GLib.PRIORITY_DEFAULT,
            None,
            self._on_write_all_async_finished,
            data)

    def _on_write_all_async_finished(self, stream, result, data):
        try:
            stream.write_all_finish(result)
        except GLib.Error as error:
            quark = GLib.quark_try_string('g-tls-error-quark')
            if error.matches(quark, Gio.TlsError.BAD_CERTIFICATE):
                self._write_stanza_buffer = None
                return

            if self._output_closed:
                self._check_for_shutdown()
                return

            self._log.error('Write Error: %s', error)
            return

        except RuntimeError as error:
            # PyGObject raises a RuntimeError when it fails to convert the
            # GError. Why it failed is printed by PyGObject
            self._log.error(error)
            return

        data = data.decode()
        self._log_stanza(data, received=False)

        if data == ' ':
            # keepalive whitespace
            self._renew_keepalive_timer()

        else:
            for stanza in self._write_stanza_buffer:
                try:
                    self.notify('data-sent', stanza)
                except Exception:
                    self._log.exception('Error while executing data-sent:')

        if self._output_closed and not self._write_queue:
            self._check_for_shutdown()
            return

        if self._write_queue:
            self._write_stanzas()

    def send(self, stanza, now=False):
        if self._state in (TCPState.DISCONNECTED, TCPState.DISCONNECTING):
            self._log.warning('send() not possible in state: %s', self._state)
            return

        if now:
            self._write_queue.appendleft(stanza)
        else:
            self._write_queue.append(stanza)

        if not self._con.get_output_stream().has_pending():
            self._write_stanzas()

    def disconnect(self):
        self._remove_keepalive_timer()
        if self.state == TCPState.CONNECTING:
            self.state = TCPState.DISCONNECTING
            self._connect_cancellable.cancel()
            return

        if self._state in (TCPState.DISCONNECTED, TCPState.DISCONNECTING):
            self._log.warning('Called disconnect on state: %s', self._state)
            return

        self.state = TCPState.DISCONNECTING
        self._finalize('disconnected')

    def _check_for_shutdown(self):
        if self._input_closed and self._output_closed:
            self._finalize('disconnected')

    def shutdown_input(self):
        self._remove_keepalive_timer()
        self._log.info('Shutdown input')
        self._input_closed = True
        self._read_cancellable.cancel()
        self._check_for_shutdown()

    def shutdown_output(self):
        self._remove_keepalive_timer()
        self.state = TCPState.DISCONNECTING
        self._log.info('Shutdown output')
        self._output_closed = True

    def _finalize(self, signal_name):
        self._remove_keepalive_timer()
        if self._con is not None:
            try:
                self._con.get_socket().shutdown(True, True)
            except GLib.Error as error:
                self._log.info(error)
        self._input_closed = True
        self._output_closed = True
        self.state = TCPState.DISCONNECTED
        self.notify(signal_name)
        self.destroy()

    def destroy(self):
        super().destroy()
        self._con = None
        self._client = None
        self._write_queue = None



# Copyright (C) 2018 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; If not, see <http://www.gnu.org/licenses/>.

import logging

from nbxmpp.namespaces import Namespace
from nbxmpp.structs import StanzaHandler
from nbxmpp.modules.date_and_time import parse_datetime
from nbxmpp.modules.base import BaseModule

log = logging.getLogger('nbxmpp.m.delay')


class Delay(BaseModule):
    def __init__(self, client):
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = [
            StanzaHandler(name='message',
                          callback=self._process_message_delay,
                          ns=Namespace.DELAY2,
                          priority=15),
            StanzaHandler(name='presence',
                          callback=self._process_presence_delay,
                          ns=Namespace.DELAY2,
                          priority=15)
        ]

    def _process_message_delay(self, _client, stanza, properties):
        if properties.is_muc_subject:
            # MUC Subjects can have a delay timestamp
            # to indicate when the user has set the subject,
            # the 'from' attr on these delays is the MUC server
            # but we treat it as user timestamp
            jids = [properties.jid.bare,
                    properties.jid.domain]

            properties.user_timestamp = parse_delay(stanza, from_=jids)

        else:
            if properties.from_muc:
                # Some servers use the MUC JID, others the domain
                jids = [properties.jid.bare,
                        properties.jid.domain]
            else:
                jids = [self._client.get_bound_jid().domain]

            server_delay = parse_delay(stanza, from_=jids)
            if server_delay is not None:
                properties.has_server_delay = True
                properties.timestamp = server_delay

            properties.user_timestamp = parse_delay(stanza, not_from=jids)

    @staticmethod
    def _process_presence_delay(_client, stanza, properties):
        properties.user_timestamp = parse_delay(stanza)


def parse_delay(stanza, epoch=True, convert='utc', from_=None, not_from=None):
    '''
    Returns the first valid delay timestamp that matches

    :param epoch:      Returns the timestamp as epoch

    :param convert:    Converts the timestamp to either utc or local

    :param from_:      Matches only delays that have the according
                       from attr set

    :param not_from:   Matches only delays that have the according
                       from attr not set
    '''
    delays = stanza.getTags('delay', namespace=Namespace.DELAY2)

    for delay in delays:
        stamp = delay.getAttr('stamp')
        if stamp is None:
            log.warning('Invalid timestamp received: %s', stamp)
            log.warning(stanza)
            continue

        delay_from = delay.getAttr('from')
        if from_ is not None:
            if delay_from not in from_:
                continue
        if not_from is not None:
            if delay_from in not_from:
                continue

        timestamp = parse_datetime(stamp, check_utc=True,
                                   epoch=epoch, convert=convert)
        if timestamp is None:
            log.warning('Invalid timestamp received: %s', stamp)
            log.warning(stanza)
            continue

        return timestamp



# Copyright (C) 2020 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=E1120

import logging

from typing import List
from typing import Optional
from typing import Set

from dataclasses import dataclass
from dataclasses import field

from nbxmpp.simplexml import Node
from nbxmpp.namespaces import Namespace

from nbxmpp.errors import MalformedStanzaError
from nbxmpp.task import iq_request_task
from nbxmpp.modules.base import BaseModule
from nbxmpp.modules.util import raise_if_error
from nbxmpp.modules.util import finalize


log = logging.getLogger('nbxmpp.m.vcard4')


ALLOWED_SEX_VALUES = ['M', 'F', 'O', 'N', 'U']
ALLOWED_KIND_VALUES = ['individual', 'group', 'org', 'location']

# Cardinality
# 1     Exactly one instance per vCard MUST be present.
# *1    Exactly one instance per vCard MAY be present.
# 1*    One or more instances per vCard MUST be present.
# *     One or more instances per vCard MAY be present.

PROPERTY_DEFINITION = {
    'source': (['altid', 'pid', 'pref', 'mediatype'], '*'),
    'kind': ([], '*1'),
    'fn': (['language', 'altid', 'pid', 'pref', 'type'], '1*'),
    'n': (['language', 'altid', 'sort-as'], '*1'),
    'nickname': (['language', 'altid', 'pid', 'pref', 'type'], '*'),
    'photo': (['altid', 'pid', 'pref', 'type', 'mediatype'], '*'),
    'bday': (['altid', 'calscale'], '*1'),
    'anniversary': (['altid', 'calscale'], '*1'),
    'gender': ([], '*1'),
    'adr': (['language', 'altid', 'pid', 'pref', 'type', 'geo', 'tz', 'label'], '*'),
    'tel': (['altid', 'pid', 'pref', 'type', 'mediatype'], '*'),
    'email': (['altid', 'pid', 'pref', 'type'], '*'),
    'impp': (['altid', 'pid', 'pref', 'type', 'mediatype'], '*'),
    'lang': (['altid', 'pid', 'pref', 'type'], '*'),
    'tz': (['altid', 'pid', 'pref', 'type', 'mediatype'], '*'),
    'geo': (['altid', 'pid', 'pref', 'type', 'mediatype'], '*'),
    'title': (['language', 'altid', 'pid', 'pref', 'type'], '*'),
    'role': (['language', 'altid', 'pid', 'pref', 'type'], '*'),
    'logo': (['language', 'altid', 'pid', 'pref', 'type', 'mediatype'], '*'),
    'org': (['language', 'altid', 'pid', 'pref', 'type', 'sort-as'], '*'),
    'member': (['altid', 'pid', 'pref', 'mediatype'], '*'),
    'related': (['altid', 'pid', 'pref', 'type', 'mediatype'], '*'),
    'categories': (['altid', 'pid', 'pref', 'type'], '*'),
    'note': (['language', 'altid', 'pid', 'pref', 'type'], '*'),
    'prodid': ([], '*1'),
    'rev': ([], '*1'),
    'sound': (['language', 'altid', 'pid', 'pref', 'type', 'mediatype'], '*'),
    'uid': ([], '*1'),
    'clientpidmap': ([], '*'),
    'url': (['altid', 'pid', 'pref', 'type', 'mediatype'], '*'),
    'key': (['altid', 'pid', 'pref', 'type', 'mediatype'], '*'),
    'fburl': (['altid', 'pid', 'pref', 'type', 'mediatype'], '*'),
    'caladruri': (['altid', 'pid', 'pref', 'type', 'mediatype'], '*'),
    'caluri': (['altid', 'pid', 'pref', 'type', 'mediatype'], '*'),
}


PROPERTY_VALUE_TYPES = {
    'bday': ['date', 'time', 'date-time', 'text'],
    'anniversary': ['date', 'time', 'date-time', 'text'],
    'key': ['text', 'uri'],
    'tel': ['uri', 'text'],
    'tz': ['text', 'uri', 'utc-offset'],
    'related': ['text', 'uri'],
}


def get_data_from_children(node, child_name):
    values = []
    child_nodes = node.getTags(child_name)
    for child_node in child_nodes:
        child_value = child_node.getData()
        if child_value:
            values.append(child_value)
    return values


def add_children(node, child_name, values):
    for value in values:
        node.addChild(child_name, payload=value)


def get_multiple_type_value(node, types):
    for type_ in types:
        value = node.getTagData(type_)
        if value:
            return type_, value

    raise ValueError('no value found')


def make_parameters(parameters):
    parameters_node = Node('parameters')
    for parameter in parameters.values():
        parameters_node.addChild(node=parameter.to_node())
    return parameters_node


def get_parameters(node):
    name = node.getName()
    definition = PROPERTY_DEFINITION[name]
    allowed_parameters = definition[0]
    parameters_node = node.getTag('parameters')
    if parameters_node is None:
        return Parameters()

    parameters = {}
    for parameter in allowed_parameters:
        parameter_node = parameters_node.getTag(parameter)
        if parameter_node is None:
            continue

        parameter_class = PARAMETER_CLASSES.get(parameter_node.getName())
        if parameter_class is None:
            continue

        parameter = parameter_class.from_node(parameter_node)
        parameters[parameter.name] = parameter

    return Parameters(parameters)


@dataclass
class Parameter:

    name: str
    type: str
    value: str

    @classmethod
    def from_node(cls, node):
        name = node.getName()
        if name != cls.name:
            raise ValueError(f'invalid parameter name: {name}')

        value = node.getTagData(cls.type)
        if not value:
            raise ValueError('no parameter value found')

        return cls(value)

    def to_node(self):
        node = Node(self.name)
        node.addChild(self.type, payload=self.value)
        return node

    def copy(self):
        return self.__class__(value=self.value)


@dataclass
class MultiParameter:

    name: str
    type: str
    values: Set[str]

    @classmethod
    def from_node(cls, node):
        name = node.getName()
        if name != cls.name:
            raise ValueError(f'invalid parameter name: {name}')

        value_nodes = node.getTags(cls.type)
        if not value_nodes:
            raise ValueError('no parameter value found')

        values = set()
        for value_node in value_nodes:
            value = value_node.getData()
            if value:
                values.add(value)

        if not values:
            raise ValueError('no parameter value found')

        return cls(values)

    def to_node(self):
        node = Node(self.name)
        for value in self.values:
            node.addChild(self.type, payload=value)
        return node

    def copy(self):
        return self.__class__(values=set(self.values))


@dataclass
class LanguageParameter(Parameter):

    name: str = field(default='language', init=False)
    type: str = field(default='language-tag', init=False)


@dataclass
class PrefParameter(Parameter):

    name: str = field(default='pref', init=False)
    type: str = field(default='integer', init=False)


@dataclass
class AltidParameter(Parameter):

    name: str = field(default='altid', init=False)
    type: str = field(default='text', init=False)


@dataclass
class PidParameter(MultiParameter):

    name: str = field(default='pid', init=False)
    type: str = field(default='text', init=False)


@dataclass
class TypeParameter(MultiParameter):

    name: str = field(default='type', init=False)
    type: str = field(default='text', init=False)


@dataclass
class MediatypeParameter(Parameter):

    name: str = field(default='mediatype', init=False)
    type: str = field(default='text', init=False)


@dataclass
class CalscaleParameter(Parameter):

    name: str = field(default='calscale', init=False)
    type: str = field(default='text', init=False)


@dataclass
class SortasParameter(MultiParameter):

    name: str = field(default='sort-as', init=False)
    type: str = field(default='text', init=False)


@dataclass
class GeoParameter(Parameter):

    name: str = field(default='geo', init=False)
    type: str = field(default='uri', init=False)


@dataclass
class TzParameter:

    name: str = field(default='tz', init=False)
    value_type: str
    value: str

    @classmethod
    def from_node(cls, node):
        name = node.getName()
        if name != cls.name:
            raise ValueError(f'invalid property name: {name}')

        value_type, value = get_multiple_type_value(node, ['text', 'uri'])
        return cls(value_type, value)

    def to_node(self):
        node = Node(self.name)
        node.addChild(self.value_type, payload=self.value)
        return node

    def copy(self):
        return self.__class__(value_type=self.value_type,
                              value=self.value)


class Parameters:
    def __init__(self, parameters=None):
        if parameters is None:
            parameters = {}
        self._parameters = parameters

    def values(self):
        return self._parameters.values()

    def get_types(self):
        parameter = self._parameters.get('type')
        if parameter is None:
            return set()

        return parameter.values

    def remove_types(self, types):
        parameter = self._parameters.get('type')
        if parameter is None:
            raise ValueError('no type parameter')

        for type_ in types:
            parameter.values.discard(type_)

        if not parameter.values:
            self._parameters.pop('type')

    def add_types(self, types):
        parameter = self._parameters.get('type')
        if parameter is None:
            parameter = TypeParameter(set(types))
            self._parameters['type'] = parameter
            return

        parameter.values.update(types)

    def copy(self):
        parameters = {}
        for name, parameter in self._parameters.items():
            parameters[name] = parameter.copy()
        return self.__class__(parameters=parameters)


PARAMETER_CLASSES = {
    'language': LanguageParameter,
    'pref': PrefParameter,
    'altid': AltidParameter,
    'pid': PidParameter,
    'type': TypeParameter,
    'mediatype': MediatypeParameter,
    'calscale': CalscaleParameter,
    'sort-as': SortasParameter,
    'geo': GeoParameter,
    'tz': TzParameter
}


@dataclass
class UriProperty:

    name: str
    value: str
    parameters: Parameters = field(default_factory=Parameters)

    @classmethod
    def from_node(cls, node):
        name = node.getName()
        if name != cls.name:
            raise ValueError(f'invalid property name: {name}')

        value = node.getTagData('uri')
        if not value:
            raise ValueError('no value found')

        parameters = get_parameters(node)

        return cls(value, parameters)

    def to_node(self):
        node = Node(self.name)
        if self.parameters:
            node.addChild(node=make_parameters(self.parameters))
        node.addChild('uri', payload=self.value)
        return node

    @property
    def is_empty(self):
        return not self.value

    def copy(self):
        return self.__class__(value=self.value,
                              parameters=self.parameters.copy())


@dataclass
class TextProperty:

    name: str
    value: str
    parameters: Parameters = field(default_factory=Parameters)

    @classmethod
    def from_node(cls, node):
        name = node.getName()
        if name != cls.name:
            raise ValueError(f'invalid property name: {name}')

        text = node.getTagData('text')
        if not text:
            raise ValueError('no value found')

        parameters = get_parameters(node)

        return cls(text, parameters)

    def to_node(self):
        node = Node(self.name)
        if self.parameters:
            node.addChild(node=make_parameters(self.parameters))
        node.addChild('text', payload=self.value)
        return node

    @property
    def is_empty(self):
        return not self.value

    def copy(self):
        return self.__class__(value=self.value,
                              parameters=self.parameters.copy())


@dataclass
class TextListProperty:

    name: str
    values: List[str]
    parameters: Parameters = field(default_factory=Parameters)

    @classmethod
    def from_node(cls, node):
        name = node.getName()
        if name != cls.name:
            raise ValueError(f'invalid property name: {name}')

        text_nodes = node.getTags('text')
        if not text_nodes:
            raise ValueError('no value found')

        values = get_data_from_children(node, 'text')
        if not values:
            raise ValueError('no values found')

        parameters = get_parameters(node)

        return cls(values, parameters)

    def to_node(self):
        node = Node(self.name)
        if self.parameters:
            node.addChild(node=make_parameters(self.parameters))
        add_children(node, 'text', self.values)
        return node

    @property
    def is_empty(self):
        return not self.values

    def copy(self):
        return self.__class__(values=list(self.values),
                              parameters=self.parameters.copy())


@dataclass
class MultipleValueProperty:

    name: str
    value_type: str
    value: str
    parameters: Parameters = field(default_factory=Parameters)

    @classmethod
    def from_node(cls, node):
        name = node.getName()
        if name != cls.name:
            raise ValueError(f'invalid property name: {name}')

        types = PROPERTY_VALUE_TYPES[cls.name]
        value_type, value = get_multiple_type_value(node, types)

        parameters = get_parameters(node)

        return cls(value_type, value, parameters)

    def to_node(self):
        node = Node(self.name)
        if self.parameters:
            node.addChild(node=make_parameters(self.parameters))
        node.addChild(self.value_type, payload=self.value)
        return node

    @property
    def is_empty(self):
        return not self.value

    def copy(self):
        return self.__class__(value_type=self.value_type,
                              value=self.value,
                              parameters=self.parameters.copy())


@dataclass
class SourceProperty(UriProperty):

    name: str = field(default='source', init=False)


@dataclass
class KindProperty(TextProperty):

    name: str = field(default='kind', init=False)

    @classmethod
    def from_node(cls, node):
        name = node.getName()
        if name != cls.name:
            raise ValueError(f'invalid property name: {name}')

        text = node.getTagData('text')
        if not text:
            raise ValueError('no value found')

        if text not in ALLOWED_KIND_VALUES:
            text = 'individual'

        parameters = get_parameters(node)

        return cls(text, parameters)


@dataclass
class FnProperty(TextProperty):

    name: str = field(default='fn', init=False)


@dataclass
class NProperty:

    name: str = field(default='n', init=False)
    surname: List[str] = field(default_factory=list)
    given: List[str] = field(default_factory=list)
    additional: List[str] = field(default_factory=list)
    prefix: List[str] = field(default_factory=list)
    suffix: List[str] = field(default_factory=list)
    parameters: Parameters = field(default_factory=Parameters)

    @classmethod
    def from_node(cls, node):
        name = node.getName()
        if name != cls.name:
            raise ValueError(f'invalid property name: {name}')

        surname = get_data_from_children(node, 'surname')
        given = get_data_from_children(node, 'given')
        additional = get_data_from_children(node, 'additional')
        prefix = get_data_from_children(node, 'prefix')
        suffix = get_data_from_children(node, 'suffix')

        parameters = get_parameters(node)

        return cls(surname, given, additional, prefix, suffix, parameters)

    def to_node(self):
        node = Node(self.name)
        if self.parameters:
            node.addChild(node=make_parameters(self.parameters))
        add_children(node, 'surname', self.surname)
        add_children(node, 'given', self.given)
        add_children(node, 'additional', self.additional)
        add_children(node, 'prefix', self.prefix)
        add_children(node, 'suffix', self.suffix)
        return node

    @property
    def is_empty(self):
        if (self.surname or
                self.given or
                self.additional or
                self.suffix or
                self.prefix):
            return False
        return True

    def copy(self):
        return self.__class__(surname=list(self.surname),
                              given=list(self.given),
                              additional=list(self.additional),
                              prefix=list(self.prefix),
                              suffix=list(self.suffix),
                              parameters=self.parameters.copy())

@dataclass
class NicknameProperty(TextListProperty):

    name: str = field(default='nickname', init=False)


@dataclass
class PhotoProperty(UriProperty):

    name: str = field(default='photo', init=False)


@dataclass
class BDayProperty(MultipleValueProperty):

    name: str = field(default='bday', init=False)


@dataclass
class AnniversaryProperty(MultipleValueProperty):

    name: str = field(default='anniversary', init=False)


@dataclass
class GenderProperty:

    name: str = field(default='gender', init=False)
    sex: Optional[str] = None
    identity: Optional[str] = None
    parameters: Parameters = field(default_factory=Parameters)

    @classmethod
    def from_node(cls, node):
        name = node.getName()
        if name != cls.name:
            raise ValueError(f'invalid property name: {name}')

        sex = node.getTagData('sex')
        if sex not in ALLOWED_SEX_VALUES:
            sex = None

        identity = node.getTagData('identity')
        if not identity:
            identity = None

        parameters = get_parameters(node)

        return cls(sex, identity, parameters)

    def to_node(self):
        node = Node(self.name)
        if self.parameters:
            node.addChild(node=make_parameters(self.parameters))
        if self.sex:
            node.addChild('sex', payload=self.sex)
        if self.identity:
            node.addChild('identity', payload=self.identity)
        return node

    @property
    def is_empty(self):
        if self.sex or self.identity:
            return False
        return True

    def copy(self):
        return self.__class__(sex=self.sex,
                              identity=self.identity,
                              parameters=self.parameters.copy())


@dataclass
class AdrProperty:

    name: str = field(default='adr', init=False)
    pobox: List[str] = field(default_factory=list)
    ext: List[str] = field(default_factory=list)
    street: List[str] = field(default_factory=list)
    locality: List[str] = field(default_factory=list)
    region: List[str] = field(default_factory=list)
    code: List[str] = field(default_factory=list)
    country: List[str] = field(default_factory=list)
    parameters: Parameters = field(default_factory=Parameters)

    @classmethod
    def from_node(cls, node):
        name = node.getName()
        if name != cls.name:
            raise ValueError(f'invalid property name: {name}')

        pobox = get_data_from_children(node, 'pobox')
        ext = get_data_from_children(node, 'ext')
        street = get_data_from_children(node, 'street')
        locality = get_data_from_children(node, 'locality')
        region = get_data_from_children(node, 'region')
        code = get_data_from_children(node, 'code')
        country = get_data_from_children(node, 'country')

        parameters = get_parameters(node)

        return cls(pobox, ext, street, locality,
                   region, code, country, parameters)

    def to_node(self):
        node = Node(self.name)
        if self.parameters:
            node.addChild(node=make_parameters(self.parameters))
        add_children(node, 'pobox', self.pobox)
        add_children(node, 'ext', self.ext)
        add_children(node, 'street', self.street)
        add_children(node, 'locality', self.locality)
        add_children(node, 'region', self.region)
        add_children(node, 'code', self.code)
        add_children(node, 'country', self.country)
        return node

    @property
    def is_empty(self):
        # pylint: disable=too-many-boolean-expressions
        if (self.pobox or
                self.ext or
                self.street or
                self.locality or
                self.region or
                self.code or
                self.country):
            return False
        return True

    def copy(self):
        return self.__class__(pobox=list(self.pobox),
                              ext=list(self.ext),
                              street=list(self.street),
                              locality=list(self.locality),
                              region=list(self.region),
                              code=list(self.code),
                              country=list(self.country),
                              parameters=self.parameters.copy())


@dataclass
class TelProperty(MultipleValueProperty):

    name: str = field(default='tel', init=False)


@dataclass
class EmailProperty(TextProperty):

    name: str = field(default='email', init=False)


@dataclass
class ImppProperty(UriProperty):

    name: str = field(default='impp', init=False)


@dataclass
class LangProperty:

    name: str = field(default='lang', init=False)
    value: str
    parameters: Parameters = field(default_factory=Parameters)

    @classmethod
    def from_node(cls, node):
        name = node.getName()
        if name != cls.name:
            raise ValueError(f'invalid property name: {name}')

        value = node.getTagData('language-tag')
        if not value:
            raise ValueError('no value found')

        parameters = get_parameters(node)

        return cls(value, parameters)

    def to_node(self):
        node = Node(self.name)
        if self.parameters:
            node.addChild(node=make_parameters(self.parameters))
        node.addChild('language-tag', payload=self.value)
        return node

    @property
    def is_empty(self):
        return not self.value

    def copy(self):
        return self.__class__(value=self.value,
                              parameters=self.parameters.copy())


@dataclass
class TzProperty(MultipleValueProperty):

    name: str = field(default='tz', init=False)


@dataclass
class GeoProperty(UriProperty):

    name: str = field(default='geo', init=False)


@dataclass
class TitleProperty(TextProperty):

    name: str = field(default='title', init=False)


@dataclass
class RoleProperty(TextProperty):

    name: str = field(default='role', init=False)


@dataclass
class LogoProperty(UriProperty):

    name: str = field(default='logo', init=False)


@dataclass
class OrgProperty(TextListProperty):

    name: str = field(default='org', init=False)


@dataclass
class MemberProperty(UriProperty):

    name: str = field(default='member', init=False)


@dataclass
class RelatedProperty(MultipleValueProperty):

    name: str = field(default='related', init=False)


@dataclass
class CategoriesProperty(TextListProperty):

    name: str = field(default='categories', init=False)


@dataclass
class NoteProperty(TextProperty):

    name: str = field(default='note', init=False)


@dataclass
class ProdidProperty(TextProperty):

    name: str = field(default='prodid', init=False)


@dataclass
class RevProperty(TextProperty):

    name: str = field(default='rev', init=False)

    @classmethod
    def from_node(cls, node):
        name = node.getName()
        if name != cls.name:
            raise ValueError(f'invalid property name: {name}')

        timestamp = node.getTagData('timestamp')
        if not timestamp:
            raise ValueError('no value found')

        parameters = get_parameters(node)

        return cls(timestamp, parameters)

    def to_node(self):
        node = Node(self.name)
        if self.parameters:
            node.addChild(node=make_parameters(self.parameters))
        node.addChild('timestamp', payload=self.value)
        return node


@dataclass
class SoundProperty(UriProperty):

    name: str = field(default='sound', init=False)


@dataclass
class UidProperty(UriProperty):

    name: str = field(default='uid', init=False)


@dataclass
class ClientpidmapProperty:

    name: str = field(default='clientpidmap', init=False)
    sourceid: int
    uri: str
    parameters: Parameters = field(default_factory=Parameters)

    @classmethod
    def from_node(cls, node):
        name = node.getName()
        if name != cls.name:
            raise ValueError(f'invalid property name: {name}')

        sourceid = node.getTagData('sourceid')
        if not sourceid:
            raise ValueError('no value found')

        uri = node.getTagData('uri')
        if not uri:
            raise ValueError('no value found')

        parameters = get_parameters(node)

        return cls(sourceid, uri, parameters)

    def to_node(self):
        node = Node(self.name)
        if self.parameters:
            node.addChild(node=make_parameters(self.parameters))
        node.addChild('sourceid', payload=self.sourceid)
        node.addChild('uri', payload=self.uri)
        return node

    @property
    def is_empty(self):
        return not self.uri

    def copy(self):
        return self.__class__(sourceid=self.sourceid,
                              uri=self.uri,
                              parameters=self.parameters.copy())


@dataclass
class UrlProperty(UriProperty):

    name: str = field(default='url', init=False)


@dataclass
class KeyProperty(MultipleValueProperty):

    name: str = field(default='key', init=False)


@dataclass
class FBurlProperty(UriProperty):

    name: str = field(default='fburl', init=False)


@dataclass
class CaladruriProperty(UriProperty):

    name: str = field(default='caladruri', init=False)


@dataclass
class CaluriProperty(UriProperty):

    name: str = field(default='calurl', init=False)


PROPERTY_CLASSES = {
    'source': SourceProperty,
    'kind': KindProperty,
    'fn': FnProperty,
    'n': NProperty,
    'nickname': NicknameProperty,
    'photo': PhotoProperty,
    'bday': BDayProperty,
    'anniversary': AnniversaryProperty,
    'gender': GenderProperty,
    'adr': AdrProperty,
    'tel': TelProperty,
    'email': EmailProperty,
    'impp': ImppProperty,
    'lang': LangProperty,
    'tz': TzProperty,
    'geo': GeoProperty,
    'title': TitleProperty,
    'role': RoleProperty,
    'logo': LogoProperty,
    'org': OrgProperty,
    'member': MemberProperty,
    'related': RelatedProperty,
    'categories': CategoriesProperty,
    'note': NoteProperty,
    'prodid': ProdidProperty,
    'rev': RevProperty,
    'sound': SoundProperty,
    'uid': UidProperty,
    'clientpidmap': ClientpidmapProperty,
    'url': UrlProperty,
    'key': KeyProperty,
    'fburl': FBurlProperty,
    'caladruri': CaladruriProperty,
    'caluri': CaluriProperty,
}


def get_property_from_name(name, node):
    property_class = PROPERTY_CLASSES.get(name)
    if property_class is None:
        return None

    try:
        return property_class.from_node(node)
    except Exception as error:
        log.warning('invalid vcard property: %s %s', error, node)
        return None


class VCard:
    def __init__(self, properties=None):
        if properties is None:
            properties = []
        self._properties = properties

    @classmethod
    def from_node(cls, node):
        properties = []
        for child in node.getChildren():
            child_name = child.getName()

            if child_name == 'group':
                group_name = child.getAttr('name')
                if not group_name:
                    continue

                group_properties = []
                for group_child in child.getChildren():
                    group_child_name = group_child.getName()
                    property_ = get_property_from_name(group_child_name,
                                                       group_child)
                    if property_ is None:
                        continue
                    group_properties.append(property_)

                properties.append((group_name, group_properties))

            else:

                property_ = get_property_from_name(child_name, child)
                if property_ is None:
                    continue
                properties.append((None, property_))

        return cls(properties)

    def to_node(self):
        vcard = Node(f'{Namespace.VCARD4} vcard')
        for group, props in self._properties:
            if group is None:
                vcard.addChild(node=props.to_node())
            else:
                group = Node(group)
                for prop in props:
                    group.addChild(node=prop.to_node())
                vcard.addChild(node=group)
        return vcard

    def get_properties(self):
        properties = []
        for group, props in self._properties:
            if group is None:
                properties.append(props)
            else:
                properties.extend(props)
        return properties

    def add_property(self, name, *args, **kwargs):
        prop = PROPERTY_CLASSES.get(name)(*args, **kwargs)
        self._properties.append((None, prop))
        return prop

    def remove_property(self, prop):
        for _group, props in list(self._properties):
            if isinstance(props, list):
                if prop in props:
                    props.remove(prop)
                    return

            elif prop is props:
                self._properties.remove((None, props))
                return

        raise ValueError('prop not found in vcard')

    def copy(self):
        properties = []
        for group_name, props in self._properties:
            if group_name is None:
                properties.append((None, props.copy()))
            else:
                group_properties = [prop.copy() for prop in props]
                properties.append((group_name, group_properties))
        return self.__class__(properties=properties)


class VCard4(BaseModule):

    _depends = {
        'publish': 'PubSub',
        'request_items': 'PubSub',
    }

    def __init__(self, client):
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = []

    @iq_request_task
    def request_vcard(self, jid=None):
        task = yield

        items = yield self.request_items(Namespace.VCARD4_PUBSUB,
                                         jid=jid,
                                         max_items=1)

        raise_if_error(items)

        if not items:
            yield task.set_result(None)

        yield _get_vcard(items[0])

    @iq_request_task
    def set_vcard(self, vcard, public=False):
        task = yield

        access_model = 'open' if public else 'presence'

        options = {
            'pubsub#persist_items': 'true',
            'pubsub#access_model': access_model,
        }

        result = yield self.publish(Namespace.VCARD4_PUBSUB,
                                    vcard.to_node(),
                                    id_='current',
                                    options=options,
                                    force_node_options=True)

        yield finalize(task, result)


def _get_vcard(item):
    vcard = item.getTag('vcard', namespace=Namespace.VCARD4)
    if vcard is None:
        raise MalformedStanzaError('vcard node missing', item)

    try:
        vcard = VCard.from_node(vcard)
    except Exception as error:
        raise MalformedStanzaError('invalid vcard: %s' % error, item)

    return vcard



# Copyright (C) 2019 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; If not, see <http://www.gnu.org/licenses/>.

import json

from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Node
from nbxmpp.protocol import Iq
from nbxmpp.structs import MuclumbusItem
from nbxmpp.structs import MuclumbusResult
from nbxmpp.const import AnonymityMode
from nbxmpp.modules.dataforms import extend_form
from nbxmpp.errors import StanzaError
from nbxmpp.errors import MalformedStanzaError
from nbxmpp.task import iq_request_task
from nbxmpp.task import http_request_task
from nbxmpp.modules.base import BaseModule
from nbxmpp.modules.util import finalize


# API Documentation
# https://search.jabber.network/docs/api

class Muclumbus(BaseModule):
    def __init__(self, client):
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = []

    @iq_request_task
    def request_parameters(self, jid):
        task = yield

        response = yield _make_parameter_request(jid)
        if response.isError():
            raise StanzaError(response)

        search = response.getTag('search', namespace=Namespace.MUCLUMBUS)
        if search is None:
            raise MalformedStanzaError('search node missing', response)

        dataform = search.getTag('x', namespace=Namespace.DATA)
        if dataform is None:
            raise MalformedStanzaError('dataform node missing', response)

        self._log.info('Muclumbus parameters received')
        yield finalize(task, extend_form(node=dataform))

    @iq_request_task
    def set_search(self, jid, dataform, items_per_page=50, after=None):
        _task = yield

        response = yield _make_search_query(jid,
                                            dataform,
                                            items_per_page,
                                            after)
        if response.isError():
            raise StanzaError(response)

        result = response.getTag('result', namespace=Namespace.MUCLUMBUS)
        if result is None:
            raise MalformedStanzaError('result node missing', response)

        items = result.getTags('item')
        if not items:
            yield MuclumbusResult(first=None,
                                  last=None,
                                  max=None,
                                  end=True,
                                  items=[])

        set_ = result.getTag('set', namespace=Namespace.RSM)
        if set_ is None:
            raise MalformedStanzaError('set node missing', response)

        first = set_.getTagData('first')
        last = set_.getTagData('last')
        try:
            max_ = int(set_.getTagData('max'))
        except Exception:
            raise MalformedStanzaError('invalid max value', response)

        results = []
        for item in items:
            jid = item.getAttr('address')
            name = item.getTagData('name')
            nusers = item.getTagData('nusers')
            description = item.getTagData('description')
            language = item.getTagData('language')
            is_open = item.getTag('is-open') is not None

            try:
                anonymity_mode = AnonymityMode(
                    item.getTagData('anonymity-mode'))
            except ValueError:
                anonymity_mode = AnonymityMode.UNKNOWN
            results.append(MuclumbusItem(jid=jid,
                                         name=name or '',
                                         nusers=nusers or '',
                                         description=description or '',
                                         language=language or '',
                                         is_open=is_open,
                                         anonymity_mode=anonymity_mode))
        yield MuclumbusResult(first=first,
                              last=last,
                              max=max_,
                              end=len(items) < max_,
                              items=results)

    @http_request_task
    def set_http_search(self, uri, keywords, after=None):
        _task = yield

        search = {'keywords': keywords}
        if after is not None:
            search['after'] = after

        body = json.dumps(search).encode()

        session = self._client.http_session
        request = session.create_request()
        request.set_request_body('application/json', body)
        request.send('POST', uri)

        request = yield request

        if not request.is_complete():
            self._log.warning(request.get_error_string())
            yield MuclumbusResult(first=None,
                                  last=None,
                                  max=None,
                                  end=True,
                                  items=[])

        response_body = request.get_data()
        response = json.loads(response_body)

        result = response['result']
        items = result.get('items')
        if items is None:
            yield MuclumbusResult(first=None,
                                  last=None,
                                  max=None,
                                  end=True,
                                  items=[])

        results = []
        for item in items:
            try:
                anonymity_mode = AnonymityMode(item['anonymity_mode'])
            except (ValueError, KeyError):
                anonymity_mode = AnonymityMode.UNKNOWN

            results.append(
                MuclumbusItem(jid=item['address'],
                              name=item['name'] or '',
                              nusers=str(item['nusers'] or ''),
                              description=item['description'] or '',
                              language=item['language'] or '',
                              is_open=item['is_open'],
                              anonymity_mode=anonymity_mode))

        yield MuclumbusResult(first=None,
                              last=result['last'],
                              max=None,
                              end=not result['more'],
                              items=results)


def _make_parameter_request(jid):
    query = Iq(to=jid, typ='get')
    query.addChild(node=Node('search',
                             attrs={'xmlns': Namespace.MUCLUMBUS}))
    return query


def _make_search_query(jid, dataform, items_per_page=50, after=None):
    search = Node('search', attrs={'xmlns': Namespace.MUCLUMBUS})
    search.addChild(node=dataform)
    rsm = search.addChild('set', namespace=Namespace.RSM)
    rsm.addChild('max').setData(items_per_page)
    if after is not None:
        rsm.addChild('after').setData(after)
    query = Iq(to=jid, typ='get')
    query.addChild(node=search)
    return query



# Copyright (C) 2018 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; If not, see <http://www.gnu.org/licenses/>.

from nbxmpp.namespaces import Namespace
from nbxmpp.structs import StanzaHandler
from nbxmpp.structs import ChatMarker
from nbxmpp.modules.base import BaseModule


class ChatMarkers(BaseModule):
    def __init__(self, client):
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = [
            StanzaHandler(name='message',
                          callback=self._process_message_marker,
                          ns=Namespace.CHATMARKERS,
                          priority=15),
        ]

    def _process_message_marker(self, _client, stanza, properties):
        type_ = stanza.getTag('received', namespace=Namespace.CHATMARKERS)
        if type_ is None:
            type_ = stanza.getTag('displayed', namespace=Namespace.CHATMARKERS)
            if type_ is None:
                type_ = stanza.getTag('acknowledged',
                                      namespace=Namespace.CHATMARKERS)
                if type_ is None:
                    return

        name = type_.getName()
        id_ = type_.getAttr('id')
        if id_ is None:
            self._log.warning('Chatmarker without id')
            self._log.warning(stanza)
            return

        properties.marker = ChatMarker(name, id_)



## c14n.py
##
## Copyright (C) 2007-2008 Brendan Taylor <whateley AT gmail.com>
##
## This file is part of Gajim.
##
## Gajim is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation; version 3 only.
##
## Gajim is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with Gajim. If not, see <http://www.gnu.org/licenses/>.
##

"""
XML canonicalisation methods (for XEP-0116)
"""

def c14n(node, is_buggy):
    s = "<" + node.name
    if node.namespace:
        if not node.parent or node.parent.namespace != node.namespace:
            s += ' xmlns="%s"' % node.namespace

    sorted_attrs = sorted(node.attrs.keys())
    for key in sorted_attrs:
        if not is_buggy and key == 'xmlns':
            continue
        val = str(node.attrs[key])
        # like XMLescape() but with whitespace and without &gt;
        s += ' %s="%s"' % (key, normalise_attr(val))
    s += ">"
    cnt = 0
    if node.kids:
        for a in node.kids:
            if (len(node.data)-1) >= cnt:
                s = s + normalise_text(node.data[cnt])
            s = s + c14n(a, is_buggy)
            cnt += 1
    if (len(node.data)-1) >= cnt:
        s = s + normalise_text(node.data[cnt])
    if not node.kids and s.endswith('>'):
        s=s[:-1]+' />'
    else:
        s = s + "</" + node.name + ">"
    return s

def normalise_attr(val):
    return val.replace('&', '&amp;').replace('<', '&lt;').replace('"', '&quot;').replace('\t', '&#x9;').replace('\n', '&#xA;').replace('\r', '&#xD;')

def normalise_text(val):
    return val.replace('&', '&amp;').replace('<', '&lt;').replace('>', '&gt;').replace('\r', '&#xD;')



# Copyright (C) 2019 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; If not, see <http://www.gnu.org/licenses/>.

from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import NodeProcessed
from nbxmpp.protocol import Node
from nbxmpp.protocol import Message
from nbxmpp.util import b64decode
from nbxmpp.util import b64encode
from nbxmpp.util import from_xs_boolean
from nbxmpp.structs import StanzaHandler
from nbxmpp.structs import OMEMOMessage
from nbxmpp.structs import OMEMOBundle
from nbxmpp.modules.base import BaseModule
from nbxmpp.modules.util import finalize
from nbxmpp.modules.util import raise_if_error
from nbxmpp.task import iq_request_task
from nbxmpp.errors import MalformedStanzaError


class OMEMO(BaseModule):

    _depends = {
        'publish': 'PubSub',
        'request_items': 'PubSub',
    }

    def __init__(self, client):
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = [
            StanzaHandler(name='message',
                          callback=self._process_omemo_devicelist,
                          ns=Namespace.PUBSUB_EVENT,
                          priority=16),
            StanzaHandler(name='message',
                          callback=self._process_omemo_message,
                          ns=Namespace.OMEMO_TEMP,
                          priority=7),
        ]

    def _process_omemo_message(self, _client, stanza, properties):
        try:
            properties.omemo = _parse_omemo_message(stanza)
            self._log.info('Received message')
        except MalformedStanzaError as error:
            self._log.warning(error)
            self._log.warning(stanza)
            return

    def _process_omemo_devicelist(self, _client, stanza, properties):
        if not properties.is_pubsub_event:
            return

        if properties.pubsub_event.node != Namespace.OMEMO_TEMP_DL:
            return

        item = properties.pubsub_event.item
        if item is None:
            # Retract, Deleted or Purged
            return

        try:
            devices = _parse_devicelist(item)
        except MalformedStanzaError as error:
            self._log.warning(error)
            self._log.warning(stanza)
            raise NodeProcessed

        if not devices:
            self._log.info('Received OMEMO devicelist: %s - no devices set',
                           properties.jid)
            return

        pubsub_event = properties.pubsub_event._replace(data=devices)
        self._log.info('Received OMEMO devicelist: %s - %s',
                       properties.jid, devices)

        properties.pubsub_event = pubsub_event

    @iq_request_task
    def set_devicelist(self, devicelist=None, public=True):
        task = yield

        access_model = 'open' if public else 'presence'

        options = {
            'pubsub#persist_items': 'true',
            'pubsub#access_model': access_model,
        }

        result = yield self.publish(Namespace.OMEMO_TEMP_DL,
                                    _make_devicelist(devicelist),
                                    id_='current',
                                    options=options,
                                    force_node_options=True)

        yield finalize(task, result)

    @iq_request_task
    def request_devicelist(self, jid=None):
        task = yield

        items = yield self.request_items(Namespace.OMEMO_TEMP_DL,
                                         max_items=1,
                                         jid=jid)

        raise_if_error(items)

        if not items:
            yield task.set_result(None)

        yield _parse_devicelist(items[0])

    @iq_request_task
    def set_bundle(self, bundle, device_id, public=True):
        task = yield

        access_model = 'open' if public else 'presence'

        options = {
            'pubsub#persist_items': 'true',
            'pubsub#access_model': access_model,
        }

        result = yield self.publish(
            f'{Namespace.OMEMO_TEMP_BUNDLE}:{device_id}',
            _make_bundle(bundle),
            id_='current',
            options=options,
            force_node_options=True)

        yield finalize(task, result)

    @iq_request_task
    def request_bundle(self, jid, device_id):
        task = yield

        items = yield self.request_items(
            f'{Namespace.OMEMO_TEMP_BUNDLE}:{device_id}',
            max_items=1,
            jid=jid)

        raise_if_error(items)

        if not items:
            yield task.set_result(None)

        yield _parse_bundle(items[0], device_id)


def _parse_omemo_message(stanza):
    '''
    <message>
      <encrypted xmlns='eu.siacs.conversations.axolotl'>
        <header sid='27183'>
          <key rid='31415'>BASE64ENCODED...</key>
          <key prekey="true" rid='12321'>BASE64ENCODED...</key>
          <!-- ... -->
          <iv>BASE64ENCODED...</iv>
        </header>
        <payload>BASE64ENCODED</payload>
      </encrypted>
      <store xmlns='urn:xmpp:hints'/>
    </message>
    '''
    encrypted = stanza.getTag('encrypted', namespace=Namespace.OMEMO_TEMP)
    if encrypted is None:
        raise MalformedStanzaError('No encrypted node found', stanza)

    header = encrypted.getTag('header')
    if header is None:
        raise MalformedStanzaError('header node not found', stanza)

    try:
        sid = int(header.getAttr('sid'))
    except Exception:
        raise MalformedStanzaError('sid attr not found', stanza)

    iv_node = header.getTag('iv')
    try:
        iv = b64decode(iv_node.getData())
    except Exception as error:
        raise MalformedStanzaError('failed to decode iv: %s' % error, stanza)

    payload = None
    payload_node = encrypted.getTag('payload')
    if payload_node is not None:
        try:
            payload = b64decode(payload_node.getData())
        except Exception as error:
            raise MalformedStanzaError('failed to decode payload: %s' % error,
                                       stanza)

    key_nodes = header.getTags('key')
    if not key_nodes:
        raise MalformedStanzaError('no keys found', stanza)

    keys = {}
    for kn in key_nodes:
        rid = kn.getAttr('rid')
        if rid is None:
            raise MalformedStanzaError('rid not found', stanza)

        prekey = kn.getAttr('prekey')
        if prekey is None:
            prekey = False
        else:
            try:
                prekey = from_xs_boolean(prekey)
            except ValueError as error:
                raise MalformedStanzaError(error, stanza)

        try:
            keys[int(rid)] = (b64decode(kn.getData()), prekey)
        except Exception as error:
            raise MalformedStanzaError('failed to decode key: %s' % error,
                                       stanza)

    return OMEMOMessage(sid=sid, iv=iv, keys=keys, payload=payload)


def _parse_bundle(item, device_id):
    '''
    <item id='current'>
      <bundle xmlns='eu.siacs.conversations.axolotl'>
        <signedPreKeyPublic signedPreKeyId='1'>
          BASE64ENCODED...
        </signedPreKeyPublic>
        <signedPreKeySignature>
          BASE64ENCODED...
        </signedPreKeySignature>
        <identityKey>
          BASE64ENCODED...
        </identityKey>
        <prekeys>
          <preKeyPublic preKeyId='1'>
            BASE64ENCODED...
          </preKeyPublic>
          <preKeyPublic preKeyId='2'>
           BASE64ENCODED...
          </preKeyPublic>
          <preKeyPublic preKeyId='3'>
            BASE64ENCODED...
          </preKeyPublic>
          <!-- ... -->
        </prekeys>
      </bundle>
    </item>
    '''
    if item is None:
        raise MalformedStanzaError('No item in node found', item)

    bundle = item.getTag('bundle', namespace=Namespace.OMEMO_TEMP)
    if bundle is None:
        raise MalformedStanzaError('No bundle node found', item)

    result = {}
    signed_prekey_node = bundle.getTag('signedPreKeyPublic')
    try:
        result['spk'] = {'key': b64decode(signed_prekey_node.getData())}
    except Exception as error:
        error = 'Failed to decode signedPreKeyPublic: %s' % error
        raise MalformedStanzaError(error, item)

    signed_prekey_id = signed_prekey_node.getAttr('signedPreKeyId')
    try:
        result['spk']['id'] = int(signed_prekey_id)
    except Exception as error:
        raise MalformedStanzaError('Invalid signedPreKeyId: %s' % error, item)

    signed_signature_node = bundle.getTag('signedPreKeySignature')
    try:
        result['spk_signature'] = b64decode(signed_signature_node.getData())
    except Exception as error:
        error = 'Failed to decode signedPreKeySignature: %s' % error
        raise MalformedStanzaError(error, item)

    identity_key_node = bundle.getTag('identityKey')
    try:
        result['ik'] = b64decode(identity_key_node.getData())
    except Exception as error:
        error = 'Failed to decode IdentityKey: %s' % error
        raise MalformedStanzaError(error, item)

    prekeys = bundle.getTag('prekeys')
    if prekeys is None or not prekeys.getChildren():
        raise MalformedStanzaError('No prekeys node found', item)

    result['otpks'] = []
    for prekey in prekeys.getChildren():
        try:
            id_ = int(prekey.getAttr('preKeyId'))
        except Exception as error:
            raise MalformedStanzaError('Invalid prekey: %s' % error, item)

        try:
            key = b64decode(prekey.getData())
        except Exception as error:
            raise MalformedStanzaError(
                'Failed to decode preKeyPublic: %s' % error, item)

        result['otpks'].append({'key': key, 'id': id_})

    return OMEMOBundle(**result,
                       device_id=device_id,
                       namespace=Namespace.OMEMO_TEMP)


def _make_bundle(bundle):
    '''
    <publish node='eu.siacs.conversations.axolotl.bundles:31415'>
      <item id='current'>
        <bundle xmlns='eu.siacs.conversations.axolotl'>
          <signedPreKeyPublic signedPreKeyId='1'>
            BASE64ENCODED...
          </signedPreKeyPublic>
          <signedPreKeySignature>
            BASE64ENCODED...
          </signedPreKeySignature>
          <identityKey>
            BASE64ENCODED...
          </identityKey>
          <prekeys>
            <preKeyPublic preKeyId='1'>
              BASE64ENCODED...
            </preKeyPublic>
            <preKeyPublic preKeyId='2'>
              BASE64ENCODED...
            </preKeyPublic>
            <preKeyPublic preKeyId='3'>
              BASE64ENCODED...
            </preKeyPublic>
            <!-- ... -->
          </prekeys>
        </bundle>
      </item>
    </publish>
    '''
    bundle_node = Node('bundle', attrs={'xmlns': Namespace.OMEMO_TEMP})
    prekey_pub_node = bundle_node.addChild(
        'signedPreKeyPublic',
        attrs={'signedPreKeyId': bundle.spk['id']})
    prekey_pub_node.addData(b64encode(bundle.spk['key']))

    prekey_sig_node = bundle_node.addChild('signedPreKeySignature')
    prekey_sig_node.addData(b64encode(bundle.spk_signature))

    identity_key_node = bundle_node.addChild('identityKey')
    identity_key_node.addData(b64encode(bundle.ik))

    prekeys = bundle_node.addChild('prekeys')
    for key in bundle.otpks:
        pre_key_public = prekeys.addChild('preKeyPublic',
                                          attrs={'preKeyId': key['id']})
        pre_key_public.addData(b64encode(key['key']))
    return bundle_node


def _make_devicelist(devicelist):
    if devicelist is None:
        devicelist = []

    devicelist_node = Node('list', attrs={'xmlns': Namespace.OMEMO_TEMP})
    for device in devicelist:
        devicelist_node.addChild('device').setAttr('id', device)

    return devicelist_node


def _parse_devicelist(item):
    '''
    <items node='eu.siacs.conversations.axolotl.devicelist'>
      <item id='current'>
        <list xmlns='eu.siacs.conversations.axolotl'>
          <device id='12345' />
          <device id='4223' />
        </list>
      </item>
    </items>
    '''
    list_node = item.getTag('list', namespace=Namespace.OMEMO_TEMP)
    if list_node is None:
        raise MalformedStanzaError('No list node found', item)

    if not list_node.getChildren():
        return []

    result = []
    devices_nodes = list_node.getChildren()
    for dn in devices_nodes:
        _id = dn.getAttr('id')
        if _id:
            result.append(int(_id))

    return result


def create_omemo_message(stanza, omemo_message, store_hint=True,
                         node_whitelist=None):
    '''
    <message>
      <encrypted xmlns='eu.siacs.conversations.axolotl'>
        <header sid='27183'>
          <key rid='31415'>BASE64ENCODED...</key>
          <key prekey="true" rid='12321'>BASE64ENCODED...</key>
          <!-- ... -->
          <iv>BASE64ENCODED...</iv>
        </header>
        <payload>BASE64ENCODED</payload>
      </encrypted>
      <store xmlns='urn:xmpp:hints'/>
    </message>
    '''

    if node_whitelist is not None:
        cleanup_stanza(stanza, node_whitelist)

    encrypted = Node('encrypted', attrs={'xmlns': Namespace.OMEMO_TEMP})
    header = Node('header', attrs={'sid': omemo_message.sid})
    for rid, (key, prekey) in omemo_message.keys.items():
        attrs = {'rid': rid}
        if prekey:
            attrs['prekey'] = 'true'
        child = header.addChild('key', attrs=attrs)
        child.addData(b64encode(key))

    header.addChild('iv').addData(b64encode(omemo_message.iv))
    encrypted.addChild(node=header)

    payload = encrypted.addChild('payload')
    payload.addData(b64encode(omemo_message.payload))

    stanza.addChild(node=encrypted)

    stanza.addChild(node=Node('encryption',
                              attrs={'xmlns': Namespace.EME,
                                     'name': 'OMEMO',
                                     'namespace': Namespace.OMEMO_TEMP}))

    stanza.setBody("You received a message encrypted with "
                   "OMEMO but your client doesn't support OMEMO.")

    if store_hint:
        stanza.addChild(node=Node('store', attrs={'xmlns': Namespace.HINTS}))


def get_key_transport_message(typ, jid, omemo_message):
    message = Message(typ=typ, to=jid)

    encrypted = Node('encrypted', attrs={'xmlns': Namespace.OMEMO_TEMP})
    header = Node('header', attrs={'sid': omemo_message.sid})
    for rid, (key, prekey) in omemo_message.keys.items():
        attrs = {'rid': rid}
        if prekey:
            attrs['prekey'] = 'true'
        child = header.addChild('key', attrs=attrs)
        child.addData(b64encode(key))

    header.addChild('iv').addData(b64encode(omemo_message.iv))
    encrypted.addChild(node=header)

    message.addChild(node=encrypted)
    return message


def cleanup_stanza(stanza, node_whitelist):
    whitelisted_nodes = []
    for tag, ns in node_whitelist:
        node = stanza.getTag(tag, namespace=ns)
        if node is not None:
            whitelisted_nodes.append(node)

    for node in list(stanza.getChildren()):
        stanza.delChild(node)

    for node in whitelisted_nodes:
        stanza.addChild(node=node)



# Copyright (C) 2006-2007 Tomasz Melcer <liori AT exroot.org>
# Copyright (C) 2006-2014 Yann Leboulanger <asterix AT lagaule.org>
# Copyright (C) 2007 Stephan Erb <steve-e AT h3c.de>
# Copyright (C) 2018 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# nbxmpp is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation; version 3 only.
#
# nbxmpp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with nbxmpp. If not, see <http://www.gnu.org/licenses/>.

# XEP-0004: Data Forms

from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import JID
from nbxmpp.simplexml import Node


# exceptions used in this module
class Error(Exception):
    pass


# when we get nbxmpp.Node which we do not understand
class UnknownDataForm(Error):
    pass


# when we get nbxmpp.Node which contains bad fields
class WrongFieldValue(Error):
    pass


# helper class to change class of already existing object
class ExtendedNode(Node):
    @classmethod
    def __new__(cls, *args, **kwargs):
        if 'extend' not in kwargs.keys() or not kwargs['extend']:
            return object.__new__(cls)

        extend = kwargs['extend']
        assert issubclass(cls, extend.__class__)
        extend.__class__ = cls
        return extend


# helper to create fields from scratch
def create_field(typ, **attrs):
    ''' Helper function to create a field of given type. '''
    field = {
        'boolean': BooleanField,
        'fixed': StringField,
        'hidden': StringField,
        'text-private': StringField,
        'text-single': StringField,
        'jid-multi': JidMultiField,
        'jid-single': JidSingleField,
        'list-multi': ListMultiField,
        'list-single': ListSingleField,
        'text-multi': TextMultiField,
    }[typ](typ=typ, **attrs)
    return field


def extend_field(node):
    """
    Helper function to extend a node to field of appropriate type
    """
    # when validation (XEP-122) will go in, we could have another classes
    # like DateTimeField - so that dicts in create_field() and
    # extend_field() will be different...
    typ = node.getAttr('type')
    field = {
        'boolean': BooleanField,
        'fixed': StringField,
        'hidden': StringField,
        'text-private': StringField,
        'text-single': StringField,
        'jid-multi': JidMultiField,
        'jid-single': JidSingleField,
        'list-multi': ListMultiField,
        'list-single': ListSingleField,
        'text-multi': TextMultiField,
    }
    if typ not in field:
        typ = 'text-single'
    return field[typ](extend=node)


def extend_form(node):
    """
    Helper function to extend a node to form of appropriate type
    """
    if node.getTag('reported') is not None:
        return MultipleDataForm(extend=node)
    return SimpleDataForm(extend=node)


class DataField(ExtendedNode):
    """
    Keeps data about one field - var, field type, labels, instructions... Base
    class for different kinds of fields. Use create_field() function to
    construct one of these
    """

    def __init__(self, typ=None, var=None, value=None, label=None, desc=None,
                 required=False, options=None, extend=None):

        if extend is None:
            ExtendedNode.__init__(self, 'field')

            self.type_ = typ
            self.var = var
            if value is not None:
                self.value = value
            if label is not None:
                self.label = label
            if desc is not None:
                self.desc = desc
            self.required = required
            self.options = options

    @property
    def type_(self):
        """
        Type of field. Recognized values are: 'boolean', 'fixed', 'hidden',
        'jid-multi', 'jid-single', 'list-multi', 'list-single', 'text-multi',
        'text-private', 'text-single'. If you set this to something different,
        DataField will store given name, but treat all data as text-single
        """
        type_ = self.getAttr('type')
        if type_ is None:
            return 'text-single'
        return type_

    @type_.setter
    def type_(self, value):
        assert isinstance(value, str)
        self.setAttr('type', value)

    @property
    def var(self):
        """
        Field identifier
        """
        return self.getAttr('var')

    @var.setter
    def var(self, value):
        assert isinstance(value, str)
        self.setAttr('var', value)

    @var.deleter
    def var(self):
        self.delAttr('var')

    @property
    def label(self):
        """
        Human-readable field name
        """
        label_ = self.getAttr('label')
        if not label_:
            label_ = self.var
        return label_

    @label.setter
    def label(self, value):
        assert isinstance(value, str)
        self.setAttr('label', value)

    @label.deleter
    def label(self):
        if self.getAttr('label'):
            self.delAttr('label')

    @property
    def description(self):
        """
        Human-readable description of field meaning
        """
        return self.getTagData('desc') or ''

    @description.setter
    def description(self, value):
        assert isinstance(value, str)
        if value == '':
            del self.description
        else:
            self.setTagData('desc', value)

    @description.deleter
    def description(self):
        desc = self.getTag('desc')
        if desc is not None:
            self.delChild(desc)

    @property
    def required(self):
        """
        Controls whether this field required to fill. Boolean
        """
        return bool(self.getTag('required'))

    @required.setter
    def required(self, value):
        required = self.getTag('required')
        if required and not value:
            self.delChild(required)
        elif not required and value:
            self.addChild('required')

    @property
    def media(self):
        """
        Media data
        """
        media = self.getTag('media', namespace=Namespace.DATA_MEDIA)
        if media:
            return Media(media)
        return None

    @media.setter
    def media(self, value):
        del self.media
        self.addChild(node=value)

    @media.deleter
    def media(self):
        media = self.getTag('media')
        if media is not None:
            self.delChild(media)

    @staticmethod
    def is_valid():
        return True, ''


class Uri(Node):
    def __init__(self, uri_tag):
        Node.__init__(self, node=uri_tag)

    @property
    def type_(self):
        """
        uri type
        """
        return self.getAttr('type')

    @type_.setter
    def type_(self, value):
        self.setAttr('type', value)

    @type_.deleter
    def type_(self):
        self.delAttr('type')

    @property
    def uri_data(self):
        """
        uri data
        """
        return self.getData()

    @uri_data.setter
    def uri_data(self, value):
        self.setData(value)

    @uri_data.deleter
    def uri_data(self):
        self.setData(None)


class Media(Node):
    def __init__(self, media_tag):
        Node.__init__(self, node=media_tag)

    @property
    def uris(self):
        """
        URIs of the media element.
        """
        return map(Uri, self.getTags('uri'))

    @uris.setter
    def uris(self, value):
        del self.uris
        for uri in value:
            self.addChild(node=uri)

    @uris.deleter
    def uris(self):
        for element in self.getTags('uri'):
            self.delChild(element)


class BooleanField(DataField):
    @property
    def value(self):
        """
        Value of field. May contain True, False or None
        """
        value = self.getTagData('value')
        if value in ('0', 'false'):
            return False
        if value in ('1', 'true'):
            return True
        if value is None:
            return False  # default value is False
        raise WrongFieldValue

    @value.setter
    def value(self, value):
        self.setTagData('value', value and '1' or '0')

    @value.deleter
    def value(self):
        value = self.getTag('value')
        if value is not None:
            self.delChild(value)


class StringField(DataField):
    """
    Covers fields of types: fixed, hidden, text-private, text-single
    """

    @property
    def value(self):
        """
        Value of field. May be any string
        """
        return self.getTagData('value') or ''

    @value.setter
    def value(self, value):
        if value is None:
            value = ''
        self.setTagData('value', value)

    @value.deleter
    def value(self):
        try:
            self.delChild(self.getTag('value'))
        except ValueError:  # if there already were no value tag
            pass

    def is_valid(self):
        if not self.required:
            return True, ''
        if not self.value:
            return False, ''
        return True, ''


class ListField(DataField):
    """
    Covers fields of types: jid-multi, jid-single, list-multi, list-single
    """

    @property
    def options(self):
        """
        Options
        """
        options = []
        for element in self.getTags('option'):
            value = element.getTagData('value')
            if value is None:
                raise WrongFieldValue
            label = element.getAttr('label')
            if not label:
                label = value
            options.append((label, value))
        return options

    @options.setter
    def options(self, values):
        del self.options
        for value, label in values:
            self.addChild('option',
                          {'label': label}).setTagData('value', value)

    @options.deleter
    def options(self):
        for element in self.getTags('option'):
            self.delChild(element)

    def iter_options(self):
        for element in self.iterTags('option'):
            value = element.getTagData('value')
            if value is None:
                raise WrongFieldValue
            label = element.getAttr('label')
            if not label:
                label = value
            yield (value, label)


class ListSingleField(ListField, StringField):
    """
    Covers list-single field
    """
    def is_valid(self):
        if not self.required:
            return True, ''
        if not self.value:
            return False, ''
        return True, ''


class JidSingleField(ListSingleField):
    """
    Covers jid-single fields
    """
    def is_valid(self):
        if self.value:
            try:
                JID.from_string(self.value)
                return True, ''
            except Exception as error:
                return False, error
        if self.required:
            return False, ''
        return True, ''


class ListMultiField(ListField):
    """
    Covers list-multi fields
    """

    @property
    def values(self):
        """
        Values held in field
        """
        values = []
        for element in self.getTags('value'):
            values.append(element.getData())
        return values

    @values.setter
    def values(self, values):
        del self.values
        for value in values:
            self.addChild('value').setData(value)

    @values.deleter
    def values(self):
        for element in self.getTags('value'):
            self.delChild(element)

    def iter_values(self):
        for element in self.getTags('value'):
            yield element.getData()

    def is_valid(self):
        if not self.required:
            return True, ''
        if not self.values:
            return False, ''
        return True, ''


class JidMultiField(ListMultiField):
    """
    Covers jid-multi fields
    """
    def is_valid(self):
        if self.values:
            for value in self.values:
                try:
                    JID.from_string(value)
                except Exception as error:
                    return False, error
            return True, ''
        if self.required:
            return False, ''
        return True, ''


class TextMultiField(DataField):
    @property
    def value(self):
        """
        Value held in field
        """
        value = ''
        for element in self.iterTags('value'):
            value += '\n' + element.getData()
        return value[1:]

    @value.setter
    def value(self, value):
        del self.value
        if value == '':
            return
        for line in value.split('\n'):
            self.addChild('value').setData(line)

    @value.deleter
    def value(self):
        for element in self.getTags('value'):
            self.delChild(element)

    def is_valid(self):
        if not self.required:
            return True, ''
        if not self.value:
            return False, ''
        return True, ''


class DataRecord(ExtendedNode):
    """
    The container for data fields - an xml element which has DataField elements
    as children
    """
    def __init__(self, fields=None, associated=None, extend=None):
        self.associated = associated
        self.vars = {}
        if extend is None:
            # we have to build this object from scratch
            Node.__init__(self)

            if fields is not None:
                self.fields = fields
        else:
            # we already have nbxmpp.Node inside - try to convert all
            # fields into DataField objects
            if fields is None:
                for field in self.iterTags('field'):
                    if not isinstance(field, DataField):
                        extend_field(field)
                    self.vars[field.var] = field
            else:
                self.fields = fields

    @property
    def fields(self):
        """
        List of fields in this record
        """
        return self.getTags('field')

    @fields.setter
    def fields(self, fields):
        del self.fields
        for field in fields:
            if not isinstance(field, DataField):
                extend_field(field)
            self.addChild(node=field)
            self.vars[field.var] = field

    @fields.deleter
    def fields(self):
        for element in self.getTags('field'):
            self.delChild(element)
            self.vars.clear()

    def iter_fields(self):
        """
        Iterate over fields in this record. Do not take associated into account
        """
        for field in self.iterTags('field'):
            yield field

    def iter_with_associated(self):
        """
        Iterate over associated, yielding both our field and associated one
        together
        """
        for field in self.associated.iter_fields():
            yield self[field.var], field

    def __getitem__(self, item):
        return self.vars[item]

    def is_valid(self):
        for field in self.iter_fields():
            if not field.is_valid()[0]:
                return False
        return True

    def is_fake_form(self):
        return bool(self.vars.get('fakeform', False))


class DataForm(ExtendedNode):
    def __init__(self, type_=None, title=None, instructions=None, extend=None):
        if extend is None:
            # we have to build form from scratch
            Node.__init__(self, 'x', attrs={'xmlns': Namespace.DATA})

        if type_ is not None:
            self.type_ = type_
        if title is not None:
            self.title = title
        if instructions is not None:
            self.instructions = instructions

    @property
    def type_(self):
        """
        Type of the form. Must be one of: 'form', 'submit', 'cancel', 'result'.
        'form' - this form is to be filled in; you will be able soon to do:
        filledform = DataForm(replyto=thisform)
        """
        return self.getAttr('type')

    @type_.setter
    def type_(self, type_):
        assert type_ in ('form', 'submit', 'cancel', 'result')
        self.setAttr('type', type_)

    @property
    def title(self):
        """
        Title of the form

        Human-readable, should not contain any \\r\\n.
        """
        return self.getTagData('title')

    @title.setter
    def title(self, title):
        self.setTagData('title', title)

    @title.deleter
    def title(self):
        try:
            self.delChild('title')
        except ValueError:
            pass

    @property
    def instructions(self):
        """
        Instructions for this form

        Human-readable, may contain \\r\\n.
        """
        # TODO: the same code is in TextMultiField. join them
        value = ''
        for valuenode in self.getTags('instructions'):
            value += '\n' + valuenode.getData()
        return value[1:]

    @instructions.setter
    def instructions(self, value):
        del self.instructions
        if value == '':
            return
        for line in value.split('\n'):
            self.addChild('instructions').setData(line)

    @instructions.deleter
    def instructions(self):
        for value in self.getTags('instructions'):
            self.delChild(value)

    @property
    def is_reported(self):
        return self.getTag('reported') is not None


class SimpleDataForm(DataForm, DataRecord):
    def __init__(self, type_=None, title=None, instructions=None, fields=None,
                 extend=None):
        DataForm.__init__(self, type_=type_, title=title,
                          instructions=instructions, extend=extend)
        DataRecord.__init__(self, fields=fields, extend=self, associated=self)

    def get_purged(self):
        simple_form = SimpleDataForm(extend=self)
        del simple_form.title
        simple_form.instructions = ''
        to_be_removed = []
        for field in simple_form.iter_fields():
            if field.required:
                # add <value> if there is not
                if hasattr(field, 'value') and not field.value:
                    field.value = ''
                # Keep all required fields
                continue
            if ((hasattr(field, 'value') and
                 not field.value and
                 field.value != 0) or
                    (hasattr(field, 'values') and not field.values)):
                to_be_removed.append(field)
            else:
                del field.label
                del field.description
                del field.media
        for field in to_be_removed:
            simple_form.delChild(field)
        return simple_form


class MultipleDataForm(DataForm):
    def __init__(self, type_=None, title=None, instructions=None, items=None,
                 extend=None):
        DataForm.__init__(self, type_=type_, title=title,
                          instructions=instructions, extend=extend)
        # all records, recorded into DataRecords
        if extend is None:
            if items is not None:
                self.items = items
        else:
            # we already have nbxmpp.Node inside - try to convert all
            # fields into DataField objects
            if items is None:
                self.items = list(self.iterTags('item'))
            else:
                for item in self.getTags('item'):
                    self.delChild(item)
                self.items = items
        reported_tag = self.getTag('reported')
        self.reported = DataRecord(extend=reported_tag)

    @property
    def items(self):
        """
        A list of all records
        """
        return list(self.iter_records())

    @items.setter
    def items(self, records):
        del self.items
        for record in records:
            if not isinstance(record, DataRecord):
                DataRecord(extend=record)
            self.addChild(node=record)

    @items.deleter
    def items(self):
        for record in self.getTags('item'):
            self.delChild(record)

    def iter_records(self):
        for record in self.getTags('item'):
            yield record



# Copyright (C) 2018 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; If not, see <http://www.gnu.org/licenses/>.

from typing import Dict

from dataclasses import dataclass
import hashlib

from nbxmpp.protocol import Iq
from nbxmpp.simplexml import Node
from nbxmpp.namespaces import Namespace
from nbxmpp.structs import StanzaHandler
from nbxmpp.errors import StanzaError
from nbxmpp.task import iq_request_task
from nbxmpp.modules.base import BaseModule


class SecurityLabels(BaseModule):
    def __init__(self, client):
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = [
            StanzaHandler(name='message',
                          callback=self._process_message_security_label,
                          ns=Namespace.SECLABEL,
                          priority=15),
        ]

    def _process_message_security_label(self, _client, stanza, properties):
        security = stanza.getTag('securitylabel', namespace=Namespace.SECLABEL)
        if security is None:
            return

        try:
            security_label = SecurityLabel.from_node(security)
        except ValueError as error:
            self._log.warning(error)
            return

        properties.security_label = security_label

    @iq_request_task
    def request_catalog(self, jid):
        _task = yield

        response = yield _make_catalog_request(self._client.domain, jid)
        if response.isError():
            raise StanzaError(response)

        catalog_node = response.getTag('catalog',
                                       namespace=Namespace.SECLABEL_CATALOG)

        items = catalog_node.getTags('item')

        labels = {}
        default = None
        for item in items:
            label = item.getAttr('selector')
            if label is None:
                continue

            security = item.getTag('securitylabel',
                                   namespace=Namespace.SECLABEL)
            if security is None:
                continue

            try:
                security_label = SecurityLabel.from_node(security)
            except ValueError:
                continue

            labels[label] = security_label

            if item.getAttr('default') == 'true':
                default = label

        yield Catalog(labels=labels, default=default)


def _make_catalog_request(domain, jid):
    iq = Iq(typ='get', to=domain)
    iq.addChild(name='catalog',
                namespace=Namespace.SECLABEL_CATALOG,
                attrs={'to': jid})
    return iq


@dataclass
class Displaymarking:
    name: str
    fgcolor: str
    bgcolor: str

    def to_node(self):
        displaymarking = Node(tag='displaymarking')
        if self.fgcolor and self.fgcolor != '#000':
            displaymarking.setAttr('fgcolor', self.fgcolor)

        if self.bgcolor and self.bgcolor != '#FFF':
            displaymarking.setAttr('bgcolor', self.bgcolor)

        if self.name:
            displaymarking.setData(self.name)

        return displaymarking

    @classmethod
    def from_node(cls, node):
        return cls(name=node.getData(),
                   fgcolor=node.getAttr('fgcolor') or '#000',
                   bgcolor=node.getAttr('bgcolor') or '#FFF')


@dataclass
class SecurityLabel:
    displaymarking: Displaymarking
    label: Node

    def to_node(self):
        security = Node(tag='securitylabel',
                        attrs={'xmlns': Namespace.SECLABEL})
        if self.displaymarking is not None:
            security.addChild(node=self.displaymarking.to_node())
        security.addChild(node=self.label)
        return security

    @classmethod
    def from_node(cls, security):
        displaymarking = security.getTag('displaymarking')
        if displaymarking is not None:
            displaymarking = Displaymarking.from_node(displaymarking)

        label = security.getTag('label')
        if label is None:
            raise ValueError('label node missing')

        return cls(displaymarking=displaymarking, label=label)

    def get_label_hash(self) -> str:
        sha = hashlib.sha512()
        sha.update(str(self.label).encode())
        return sha.hexdigest()

@dataclass
class Catalog:
    labels: Dict[str, SecurityLabel]
    default: str

    def get_label_names(self):
        return list(self.labels.keys())



# Copyright (C) 2018 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; If not, see <http://www.gnu.org/licenses/>.

from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Iq
from nbxmpp.const import REGISTER_FIELDS
from nbxmpp.structs import RegisterData
from nbxmpp.errors import StanzaError
from nbxmpp.errors import MalformedStanzaError
from nbxmpp.modules.dataforms import create_field
from nbxmpp.modules.dataforms import extend_form
from nbxmpp.modules.dataforms import SimpleDataForm
from nbxmpp.modules.bits_of_binary import parse_bob_data


def _make_password_change_request(domain, username, password):
    iq = Iq('set', Namespace.REGISTER, to=domain)
    query = iq.getQuery()
    query.setTagData('username', username)
    query.setTagData('password', password)
    return iq


def _make_password_change_with_form(domain, form):
    iq = Iq('set', Namespace.REGISTER, to=domain)
    iq.setQueryPayload(form)
    return iq


def _make_register_form(jid, form):
    iq = Iq('set', Namespace.REGISTER, to=jid)
    if form.is_fake_form():
        query = iq.getTag('query')
        for field in form.iter_fields():
            if field.var == 'fakeform':
                continue
            query.addChild(field.var, payload=[field.value])
        return iq

    iq.setQueryPayload(form)
    return iq


def _make_unregister_request(jid):
    iq = Iq('set', to=jid)
    query = iq.setQuery()
    query.setNamespace(Namespace.REGISTER)
    query.addChild('remove')
    return iq


def _parse_oob_url(query):
    oob = query.getTag('x', namespace=Namespace.X_OOB)
    if oob is not None:
        return oob.getTagData('url') or None
    return None


def _parse_form(stanza):
    query = stanza.getTag('query', namespace=Namespace.REGISTER)
    form = query.getTag('x', namespace=Namespace.DATA)
    if form is None:
        return None

    form = extend_form(node=form)
    field = form.vars.get('FORM_TYPE')
    if field is None:
        return None

    # Invalid urn:xmpp:captcha used by ejabberd
    # See https://github.com/processone/ejabberd/issues/3045
    if field.value in ('jabber:iq:register', 'urn:xmpp:captcha'):
        return form
    return None


def _parse_fields_form(query):
    fields = []
    for field in query.getChildren():
        field_name = field.getName()
        if field_name not in REGISTER_FIELDS:
            continue

        required = field_name in ('username', 'password')
        typ = 'text-single' if field_name != 'password' else 'text-private'
        fields.append(create_field(typ=typ,
                                   var=field_name,
                                   required=required))

    if not fields:
        return None

    fields.append(create_field(typ='hidden', var='fakeform'))
    return SimpleDataForm(type_='form',
                          instructions=query.getTagData('instructions'),
                          fields=fields)


def _parse_register_data(response):
    query = response.getTag('query', namespace=Namespace.REGISTER)
    if query is None:
        raise StanzaError(response)

    instructions = query.getTagData('instructions') or None

    data = RegisterData(instructions=instructions,
                        form=_parse_form(response),
                        fields_form=_parse_fields_form(query),
                        oob_url=_parse_oob_url(query),
                        bob_data=parse_bob_data(query))

    if (data.form is None and
            data.fields_form is None and
            data.oob_url is None):
        raise MalformedStanzaError('invalid register response', response)

    return data



# Copyright (C) 2018 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; If not, see <http://www.gnu.org/licenses/>.

import re
import time
import logging
from datetime import datetime
from datetime import timedelta
from datetime import timezone
from datetime import tzinfo

log = logging.getLogger('nbxmpp.m.date_and_time')

PATTERN_DATETIME = re.compile(r'''
    ([0-9]{4}-[0-9]{2}-[0-9]{2})    # Date
    T                               # Separator
    ([0-9]{2}:[0-9]{2}:[0-9]{2})    # Time
    (?P<frac>\.[0-9]{0,6})?         # Fractual Seconds
    [0-9]*                          # lose everything > 6
    (Z|[-+][0-9]{2}:[0-9]{2})       # UTC Offset
    $                               # End of String
''', re.VERBOSE)

ZERO = timedelta(0)
HOUR = timedelta(hours=1)
SECOND = timedelta(seconds=1)

STDOFFSET = timedelta(seconds=-time.timezone)
if time.daylight:
    DSTOFFSET = timedelta(seconds=-time.altzone)
else:
    DSTOFFSET = STDOFFSET

DSTDIFF = DSTOFFSET - STDOFFSET


class LocalTimezone(tzinfo):
    '''
    A class capturing the platform's idea of local time.
    May result in wrong values on historical times in
    timezones where UTC offset and/or the DST rules had
    changed in the past.
    '''
    def fromutc(self, dt):
        assert dt.tzinfo is self
        stamp = (dt - datetime(1970, 1, 1, tzinfo=self)) // SECOND
        args = time.localtime(stamp)[:6]
        dst_diff = DSTDIFF // SECOND
        # Detect fold
        fold = (args == time.localtime(stamp - dst_diff))
        return datetime(*args, microsecond=dt.microsecond,
                        tzinfo=self, fold=fold)

    def utcoffset(self, dt):
        if self._isdst(dt):
            return DSTOFFSET
        return STDOFFSET

    def dst(self, dt):
        if self._isdst(dt):
            return DSTDIFF
        return ZERO

    def tzname(self, dt):
        return 'local'

    @staticmethod
    def _isdst(dt):
        tt = (dt.year, dt.month, dt.day,
              dt.hour, dt.minute, dt.second,
              dt.weekday(), 0, 0)
        stamp = time.mktime(tt)
        tt = time.localtime(stamp)
        return tt.tm_isdst > 0


def create_tzinfo(hours=0, minutes=0, tz_string=None):
    if tz_string is None:
        return timezone(timedelta(hours=hours, minutes=minutes))

    if tz_string.lower() == 'z':
        return timezone.utc

    try:
        hours, minutes = map(int, tz_string.split(':'))
    except Exception:
        log.warning('Wrong tz string: %s', tz_string)
        return None

    if hours not in range(-24, 24):
        log.warning('Wrong tz string: %s', tz_string)
        return None

    if minutes not in range(0, 59):
        log.warning('Wrong tz string: %s', tz_string)
        return None

    if hours in (24, -24) and minutes != 0:
        log.warning('Wrong tz string: %s', tz_string)
        return None
    return timezone(timedelta(hours=hours, minutes=minutes))


def parse_datetime(
    timestring: str | None,
    check_utc: bool = False,
    convert: str | None = 'utc',
    epoch: bool = False
) -> datetime | float | None:
    '''
    Parse a XEP-0082 DateTime Profile String

    :param timestring: a XEP-0082 DateTime profile formated string

    :param check_utc:  if True, returns None if timestring is not
                       a timestring expressing UTC

    :param convert:    convert the given timestring to utc or local time

    :param epoch:      if True, returns the time in epoch

    Examples:
    '2017-11-05T01:41:20Z'
    '2017-11-05T01:41:20.123Z'
    '2017-11-05T01:41:20.123+05:00'

    return a datetime or epoch
    '''
    if timestring is None:
        return None
    if convert not in (None, 'utc', 'local'):
        raise TypeError('"%s" is not a valid value for convert')

    match = PATTERN_DATETIME.match(timestring)
    if match is None:
        return None

    timestring = ''.join(match.groups(''))
    strformat = '%Y-%m-%d%H:%M:%S%z'
    if match.group('frac'):
        # Fractional second addendum to Time
        strformat = '%Y-%m-%d%H:%M:%S.%f%z'

    try:
        date_time = datetime.strptime(timestring, strformat)
    except ValueError:
        return None

    if not 1 < date_time.year < 9999:
        # Raise/Reduce MIN/MAX year so converting to different
        # timezones cannot get out of range
        return None

    if check_utc:
        if convert != 'utc':
            raise ValueError(
                'check_utc can only be used with convert="utc"')

        if date_time.tzinfo != timezone.utc:
            return None

        if epoch:
            return date_time.timestamp()
        return date_time

    if convert == 'utc':
        date_time = date_time.astimezone(timezone.utc)
        if epoch:
            return date_time.timestamp()
        return date_time

    if epoch:
        # epoch is always UTC, use convert='utc' or check_utc=True
        raise ValueError(
            'epoch not available while converting to local')

    if convert == 'local':
        date_time = date_time.astimezone(LocalTimezone())
        return date_time

    # convert=None
    return date_time


def get_local_time():
    formated_time = time.strftime('%Y-%m-%dT%H:%M:%SZ', time.gmtime())
    isdst = time.localtime().tm_isdst
    zone = -(time.timezone, time.altzone)[isdst] / 60.0
    zone = (zone / 60, abs(zone % 60))
    tzo = '%+03d:%02d' % zone
    return formated_time, tzo





from precis_i18n import get_profile


_localpart_disallowed_chars = set('"&\'/:<>@')


def enforce_precis_username(localpart: str) -> str:
    if _localpart_disallowed_chars & set(localpart):
        raise ValueError('Input contains prohibited codepoint: %s' % localpart)

    username = get_profile('UsernameCaseMapped')
    return username.enforce(localpart)


def enforce_precis_opaque(resourcepart: str) -> str:
    opaque = get_profile('OpaqueString')
    return opaque.enforce(resourcepart)



# Copyright (C) 2020 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; If not, see <http://www.gnu.org/licenses/>.

import logging

from gi.repository import Soup
from gi.repository import GLib
from gi.repository import Gio

from nbxmpp.const import TCPState
from nbxmpp.http import DEFAULT_USER_AGENT
from nbxmpp.util import get_websocket_close_string
from nbxmpp.util import convert_tls_error_flags
from nbxmpp.connection import Connection

log = logging.getLogger('nbxmpp.websocket')


class WebsocketConnection(Connection):
    def __init__(self, *args, **kwargs):
        Connection.__init__(self, *args, **kwargs)

        self._session = Soup.Session()
        self._session.set_user_agent(DEFAULT_USER_AGENT)

        if self._log.getEffectiveLevel() == logging.INFO:
            self._session.add_feature(
                Soup.Logger.new(Soup.LoggerLogLevel.BODY))

        self._websocket = None
        self._cancellable = Gio.Cancellable()

        self._input_closed = False
        self._output_closed = False

    def connect(self):
        self._log.info('Try to connect to %s', self._address.uri)

        self.state = TCPState.CONNECTING

        message = Soup.Message.new('GET', self._address.uri)
        assert message is not None
        message.connect('accept-certificate', self._check_certificate)
        message.connect('notify::tls-peer-certificate',
                        self._on_certificate_set)
        message.connect('network-event', self._on_network_event)
        message.set_flags(Soup.MessageFlags.NO_REDIRECT)
        self._session.websocket_connect_async(message,
                                              None,
                                              ['xmpp'],
                                              GLib.PRIORITY_DEFAULT,
                                              self._cancellable,
                                              self._on_connect,
                                              None)

    def _on_connect(self, session, result, _user_data):
        # TODO: check if protocol 'xmpp' is set
        try:
            self._websocket = session.websocket_connect_finish(result)
        except GLib.Error as error:
            quark = GLib.quark_try_string('g-io-error-quark')
            if error.matches(quark, Gio.IOErrorEnum.CANCELLED):
                self._finalize('disconnected')
                return

            self._log.info('Connection Error: %s', error)
            self._finalize('connection-failed')
            return

        self._websocket.set_keepalive_interval(5)
        self._websocket.set_max_incoming_payload_size(1048576)
        self._websocket.connect('message', self._on_websocket_message)
        self._websocket.connect('closed', self._on_websocket_closed)
        self._websocket.connect('closing', self._on_websocket_closing)
        self._websocket.connect('error', self._on_websocket_error)
        self._websocket.connect('pong', self._on_websocket_pong)

        self.state = TCPState.CONNECTED
        self.notify('connected')

    def start_tls_negotiation(self):
        # Soup.Session does this automatically
        raise NotImplementedError

    def _check_certificate(self,
                           _message,
                           certificate,
                           errors):

        self._peer_certificate = certificate
        self._peer_certificate_errors = convert_tls_error_flags(errors)

        if self._accept_certificate():
            return True

        self.notify('bad-certificate')
        self._cancellable.cancel()
        return False

    def _on_network_event(
        self,
        _message: Soup.Message,
        event: Gio.SocketClientEvent,
        connection: Gio.TlsConnection
    ) -> None:

        if event == Gio.SocketClientEvent.TLS_HANDSHAKED:
            self._tls_con = connection

    def _on_certificate_set(self, message, _param):
        if self._peer_certificate is not None:
            return

        # If the cert has errors _check_certificate() will set the cert.
        self._peer_certificate = message.props.tls_peer_certificate
        self._peer_certificate_errors = convert_tls_error_flags(
            message.props.tls_peer_certificate_errors)

    def _on_websocket_message(self, _websocket, _type, message):
        data = message.get_data().decode()
        self._log_stanza(data)

        if self._input_closed:
            self._log.warning('Received data after stream closed')
            return

        self.notify('data-received', data)

    def _on_websocket_pong(self, _websocket, _message):
        self._log.info('Pong received')

    def _on_websocket_closed(self, websocket):
        self._log.info('Closed %s', get_websocket_close_string(websocket))
        self._finalize('disconnected')

    def _on_websocket_closing(self, _websocket):
        self._log.info('Closing')

    def _on_websocket_error(self, _websocket, error):
        self._log.error(error)
        if self._state not in (TCPState.DISCONNECTED, TCPState.DISCONNECTING):
            self._finalize('disconnected')

    def send(self, stanza, now=False):
        if self._state in (TCPState.DISCONNECTED, TCPState.DISCONNECTING):
            self._log.warning('send() not possible in state: %s', self._state)
            return

        data = str(stanza)
        self._websocket.send_text(data)
        self._log_stanza(data, received=False)
        self.notify('data-sent', stanza)

    def disconnect(self):
        if self._state == TCPState.CONNECTING:
            self.state = TCPState.DISCONNECTING
            self._cancellable.cancel()
            return

        if self._state in (TCPState.DISCONNECTED, TCPState.DISCONNECTING):
            self._log.warning('Called disconnect on state: %s', self._state)
            return

        self._websocket.close(Soup.WebsocketCloseCode.NORMAL, None)
        self.state = TCPState.DISCONNECTING

    def _check_for_shutdown(self):
        if self._input_closed and self._output_closed:
            self._websocket.close(Soup.WebsocketCloseCode.NORMAL, None)

    def shutdown_input(self):
        self._log.info('Shutdown input')
        self._input_closed = True
        self._check_for_shutdown()

    def shutdown_output(self):
        self.state = TCPState.DISCONNECTING
        self._log.info('Shutdown output')
        self._output_closed = True

    def _finalize(self, signal_name):
        self._input_closed = True
        self._output_closed = True
        self.state = TCPState.DISCONNECTED
        self.notify(signal_name)
        self.destroy()

    def destroy(self):
        super().destroy()
        self._session.abort()
        self._session = None
        self._websocket = None



# Copyright (C) 2018 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; If not, see <http://www.gnu.org/licenses/>.

from typing import Optional

import logging
import hashlib

from nbxmpp.simplexml import Node
from nbxmpp.namespaces import Namespace
from nbxmpp.structs import BobData
from nbxmpp.util import b64decode

log = logging.getLogger('nbxmpp.m.bob')


def parse_bob_data(stanza: Node) -> Optional[BobData]:
    data_node = stanza.getTag('data', namespace=Namespace.BOB)
    if data_node is None:
        return None

    cid = data_node.getAttr('cid')
    type_ = data_node.getAttr('type')
    max_age = data_node.getAttr('max-age')
    if max_age is not None:
        try:
            max_age = int(max_age)
        except Exception:
            log.exception(stanza)
            return None

    if cid is None or type_ is None:
        log.warning('Invalid data node (no cid or type attr): %s', stanza)
        return None

    try:
        algo_hash = cid.split('@')[0]
        algo, hash_ = algo_hash.split('+')
    except Exception:
        log.exception('Invalid cid: %s', stanza)
        return None

    bob_data = data_node.getData()
    if not bob_data:
        log.warning('No bob data found: %s', stanza)
        return None

    try:
        bob_data = b64decode(bob_data)
    except Exception:
        log.warning('Unable to decode data')
        log.exception(stanza)
        return None

    try:
        sha = hashlib.new(algo)
    except ValueError as error:
        log.warning(stanza)
        log.warning(error)
        return None

    sha.update(bob_data)
    if sha.hexdigest() != hash_:
        log.warning('Invalid hash: %s', stanza)
        return None

    return BobData(algo=algo,
                   hash_=hash_,
                   max_age=max_age,
                   data=bob_data,
                   cid=cid,
                   type=type_)
